#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""Mindpet delta init"""
from mindpet.delta.adapter import AdapterDense, AdapterLayer
from mindpet.delta.lora import LoRADense
from mindpet.delta.low_rank_adapter import LowRankAdapterDense, LowRankAdapterLayer
from mindpet.delta.prefix_layer import PrefixLayer
from mindpet.delta.ptuning2 import PrefixEncoder
from mindpet.delta.r_drop import RDropLoss, rdrop_repeat

__all__ = [
    "LoRADense",
    "PrefixLayer",
    "LowRankAdapterDense",
    "LowRankAdapterLayer",
    "AdapterDense",
    "AdapterLayer",
    "RDropLoss",
    "rdrop_repeat",
    "PrefixEncoder",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module contains utility functions for freezing modules in the MindPet framework.

Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""
from mindpet.graph.freeze_utils import freeze_modules, freeze_delta, freeze_from_config
from mindpet.graph.ckpt_util import TrainableParamsCheckPoint

__all__ = ['freeze_modules', 'freeze_delta', 'freeze_from_config', 'TrainableParamsCheckPoint']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""MindPet Exceptions"""
from mindpet.utils.constants import EMPTY_STRING


# MindPet包基础异常
class MindPetError(Exception):
    def __init__(self, error_info=None):
        super().__init__()
        self.error_info = error_info

    def __str__(self):
        return self.error_info if self.error_info else EMPTY_STRING


# 基础操作异常
class UnexpectedError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


class MakeDirError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


class ReadYamlFileError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


class ManualCancelError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


# 任务异常
class CreateProcessError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


class MonitorProcessRspError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


class TaskError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


class UnsupportedPlatformError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


# 接口参数校验异常
class LinkPathError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


class LowDiskFreeSizeRiskError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


class FileOversizeError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


class PathContentError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


class PathLengthError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


class FileNameLengthError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


class AbsolutePathError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


class PathGranularityError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


class PathOwnerError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


class PathModeError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


class PathRightEscalationError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


class PathLoopError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


# model_config配置文件校验异常
class ModelConfigKeysInfoError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


class ModelConfigParamsInfoError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)


class ModelConfigFreezeInfoError(MindPetError):
    # pylint: disable=W0235
    def __init__(self, error_info=None):
        super().__init__(error_info)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""IO APIs"""
import os
import json
import yaml
from mindpet.utils.exceptions import LinkPathError
from mindpet.utils.constants import DEFAULT_FLAGS, DEFAULT_MODES, EMPTY_STRING


def is_link_path(path):
    """
    判断路径是否是软链接
    :return: 是否是软链接
    """
    path = os.path.abspath(str(path))
    return os.path.islink(path)


def read_yaml_file(path, flags=None, modes=None):
    """
    读取文件内容

    :param path: 待读取文件路径
    :param flags 文件I/O flag标识
    :param modes 文件I/O modes标识
    :return: 读取内容
    """
    if flags is None:
        flags = DEFAULT_FLAGS

    if modes is None:
        modes = DEFAULT_MODES

    if not os.path.exists(path):
        raise FileNotFoundError('File is not found.')

    if is_link_path(path):
        raise LinkPathError('Detect link path, refuse reading file.')

    with os.fdopen(os.open(path, flags, modes), 'rb') as file:
        content = yaml.safe_load(file)

    return content


def read_json_file(path, flags=None, modes=None):
    """
    读取json文件内容

    :param path: 待读取文件路径
    :param flags 文件I/O flag标识
    :param modes 文件I/O modes标识
    :return: 读取内容
    """
    if flags is None:
        flags = DEFAULT_FLAGS

    if modes is None:
        modes = DEFAULT_MODES

    if not os.path.exists(path):
        raise FileNotFoundError('File is not found.')

    if is_link_path(path):
        raise LinkPathError('Detect link path, refuse reading file.')

    if os.path.getsize(path) == 0:
        return EMPTY_STRING

    with os.fdopen(os.open(path, flags, modes), 'rb') as file:
        content = json.load(file)

    return content



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""Mindpet sdk APIs."""

from mindpet import mindpet_sdk

__all__ = ["mindpet_sdk"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""Log APIs"""
import os
import stat
import platform
from typing import Dict, List, Tuple, Union

from mindpet.utils.exceptions import MakeDirError, UnsupportedPlatformError, PathOwnerError
from mindpet.utils.constants import EMPTY_STRING

LOG_CONTENT_BLACK_LIST = ('\r', '\n', '\t', '\f', '\v', '\b', '\u000A', '\u000D', '\u000C',
                          '\u000B', '\u0008', '\u007F', '\u0009')
BLANKS = '    '


def check_list(var_name: str, list_var: Union[Tuple, List], num: int):
    """确定节点或者设备的参数的合法性

    Args:
        var_name (str): 参数名称
        list_var (tuple or list): list参数列表
        num (int): 节点数或设备数

    Returns:
        None
    """
    for value in list_var:
        if value >= num:
            raise ValueError(f'The index of the {var_name} needs to be less than the number of nodes {num}.')


def generate_rank_list(stdout_nodes: Union[List, Tuple], stdout_devices: Union[List, Tuple]):
    """生成需要输出的日志内容的rank列表

    Args:
        stdout_nodes (list or tuple): 需要输出日志内容的节点列表
        stdout_devices (list or tuple): 需要输出日志内筒的设备列表

    Returns:
        rank_list (list): 日志内容输出的rank列表
    """
    rank_list = []
    for node in stdout_nodes:
        for device in stdout_devices:
            rank_list.append(8 * node + device)

    return rank_list


def convert_nodes_devices_input(var: Union[List, Tuple, Dict[str, int], None], num: int) -> Union[List, Tuple]:
    """转换节点和设备的参数，转换为列表形式

    Args:
        var (list[int] or tuple[int] or dict[str, int] or optional):
            需要转换的参数名
        num (str): 节点数或者设备数

    Returns:
        var (list[int] or tuple[int]): list形式的节点或者设备
    """
    if var is None:
        var = tuple(range(num))
    elif isinstance(var, dict):
        var = tuple(range(var['start'], var['end']))

    return var


def create_dirs(path, mode=0o750):
    """递归创建文件夹"""
    p_path = os.path.join(path, "../..")
    abspath = os.path.abspath(p_path)
    if not os.path.exists(abspath):
        os.makedirs(abspath, mode, exist_ok=True)
    if not os.path.exists(path):
        os.makedirs(path, mode, exist_ok=True)


def get_env_info(env_name, default_value):
    env_info = os.getenv(env_name, default_value)
    try:
        res = int(env_info)
    except ValueError:
        res = default_value
    if res < 0:
        res = default_value
    return res


def get_rank_info() -> Tuple[int, int, int, int]:
    """获得环境中的rank信息

    Returns:
        server_id (int): 设备id
        device_id (int): 在本地的卡id
        rank_id (int): 全局的卡id
        rank_size (int): 总卡数
    """
    rank_id = get_env_info('RANK_ID', 0)
    rank_size = get_env_info('RANK_SIZE', 1)
    server_id = get_env_info('SERVER_ID', 0)
    device_id = get_env_info('DEVICE_ID', 0)
    device_id = device_id % 8
    rank_info = rank_id, rank_size, server_id, device_id

    return rank_info


def get_num_nodes_devices(rank_size: int) -> Tuple[int, int]:
    """根据rank大小计算节点数和设备数

    Args:
        rank_size (int): rank 大小

    Returns:
       num_nodes (int):节点数
       num_devices (int): 卡数
    """
    if rank_size in (2, 4, 8):
        num_nodes = 1
        num_devices = rank_size
    else:
        num_nodes = rank_size // 8
        num_devices = 8

    return num_nodes, num_devices


def log_args_black_list_characters_replace(args):
    """日志内容参数黑名单校验"""
    res = []
    if args is None or len(args) == 0:
        return args
    if isinstance(args, (list, tuple)):
        for arg in args:
            replace = character_replace(arg)
            res.append(replace)
        args = tuple(res)
    else:
        args = character_replace(args)
    return args


def character_replace(content):
    """字符串黑名单过滤"""
    if not isinstance(content, str):
        return content

    content = str(content)
    for forbidden_str in LOG_CONTENT_BLACK_LIST:
        if forbidden_str in content:
            content = content.replace(forbidden_str, '')
    while BLANKS in content:
        content = content.replace(BLANKS, ' ')
    return content


def wrap_local_working_directory(file_name, specific_path_config=None):
    """
    根据当前用户身份, 组装对应本地缓存文件路径
    默认文件存放位置$HOME/.cache/Huawei/mxTuningKit
    需要存放特定文件夹需指定specific_path_config = {'path': 路径, 'rule': 权限范围}
    """

    def home_path_check(path):
        """
        HOME路径属主校验以及其他用户不可写校验
        """
        path_owner = os.stat(path).st_uid

        current_login_user = os.geteuid()

        if path_owner != current_login_user:
            raise PathOwnerError('The owner of $HOME path is not current login user.')

    if file_name is None or file_name == '':
        raise ValueError('[file_name] is None or empty.')

    # 针对Linux环境, 在当前用户$HOME目录下指定文件夹存放文件, 同时约束文件夹访问权限
    if platform.system().lower() == 'linux':
        home_path = os.getenv('HOME')

        if home_path is None or home_path == EMPTY_STRING:
            raise ValueError('Wrong environment variables')

        home_path_check(home_path)

        full_path = os.path.join(home_path, '.cache', 'Huawei', 'mxTuningKit')

        if not os.path.exists(full_path):
            try:
                os.makedirs(full_path, exist_ok=True, mode=stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP)
            except Exception as ex:
                raise MakeDirError('Failed to create directory.') from ex
    else:
        raise UnsupportedPlatformError('Current platform is not supported, only support Linux.')

    if specific_path_config is not None:
        specific_path, specific_path_rule = specific_path_config_legality_check(specific_path_config)

        full_path = os.path.join(full_path, specific_path)

        if not os.path.exists(full_path):
            try:
                os.makedirs(full_path, exist_ok=True, mode=specific_path_rule)
            except Exception as ex:
                raise MakeDirError('Failed to create directory.') from ex

    wrap_path = os.path.join(full_path, file_name).replace('\\', '/')

    if not os.path.exists(wrap_path):
        try:
            os.makedirs(wrap_path, stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP, exist_ok=True)
        except Exception as ex:
            raise MakeDirError('Failed to create directory.') from ex

    return wrap_path


def specific_path_config_legality_check(specific_path_config):
    """specific_path_config_legality_check"""
    specific_path = specific_path_config.get('path')

    if specific_path is None:
        raise ValueError('[specific_path] is None.')

    specific_path_rule = specific_path_config.get('rule')

    if specific_path_rule is None:
        raise ValueError('[specific_path_rule] is None.')

    return [specific_path, specific_path_rule]


class Const:
    """Const"""
    def __init__(self):
        self.level = ('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL')
        self.local_default_log_file_dir = None
        self.node_dir_format = 'node_{}/'
        self.rank_dir_format = 'device_{}/'
        self.default_filehandler_format = \
            '[%(levelname)s] %(asctime)s [%(pid)s] [%(source)s] : %(message)s'
        self.default_stdout_format = '[%(levelname)s] %(asctime)s [%(pid)s] [%(source)s] : %(message)s'
        self.default_model_log_format = '%(asctime)s [%(pid)s] [Model]: %(message)s'

    def __setattr__(self, var, value):
        if not var.islower():
            raise ValueError(f'Const name {var} is not all lowercase.')
        self.__dict__[var] = value

    def get_local_default_log_file_dir(self):
        self.local_default_log_file_dir = wrap_local_working_directory('log')


const = Const()


def check_link_path(path):
    """
    校验传入的路径是否为软链接
    """
    path = os.path.abspath(path)
    return os.path.islink(path)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) Huawei Technologies Co., Ltd. 2022-2023, All rights reserved.
"""Adapter Cell"""
from collections import OrderedDict

from mindspore import nn
import mindspore.common.dtype as mstype
from mindspore.ops import operations as P
from mindspore.ops import functional as F
try:
    from mindspore.nn.transformer.layers import _Linear, _args_type_validator_check, _valid_value_checks
    from mindspore._checkparam import Validator
except ImportError:
    import mindspore._checkparam as Validator
    from mindformers.modules.layers import Linear, _args_type_validator_check, _valid_value_checks
    _Linear = Linear
from mindpet.delta.delta_constants import VALID_TENSOR_DATATYPE
from mindpet.utils.version_control import get_activation, _activation


class AdapterLayer(nn.Cell):
    """
    定义微调算法adapter layer层，初始化adapter layer层参数，包括矩阵参数、激活层等。

    Args:
        hidden_size (int): 隐藏层输出的特征向量维度。
        bottleneck_size (int): adapter模块中bottleneck的隐藏维度。
        non_linearity (str): adapter模块中bottleneck非线性激活函数，
            str的值引用自mindspore中 `get_activation` 方法所支持的激活函数类型.默认值：'gelu'。
        param_init_type (dtype.Number): 表示dense中模块的参数初始化类型。
        compute_dtype (dtype.Number): 表示dense中矩阵乘法的计算类型。
    """

    @_args_type_validator_check(hidden_size=Validator.check_positive_int,
                                bottleneck_size=Validator.check_positive_int,
                                non_linearity=_valid_value_checks(list(_activation.keys()),
                                                                  "AdapterLayer"),
                                param_init_type=_valid_value_checks(VALID_TENSOR_DATATYPE, "AdapterLayer"),
                                compute_dtype=_valid_value_checks(VALID_TENSOR_DATATYPE, "AdapterLayer"))

    def __init__(
            self,
            hidden_size: int,
            bottleneck_size: int = 64,
            non_linearity: str = "gelu",
            param_init_type: mstype = mstype.float32,
            compute_dtype: mstype = mstype.float16):

        super().__init__()

        self.bottleneck_size = bottleneck_size
        self.non_linearity_name = non_linearity

        adapter_dict = OrderedDict()
        adapter_dict["mindpet_delta_adapter_down_sampler"] = _Linear(hidden_size,
                                                                bottleneck_size,
                                                                compute_dtype=compute_dtype,
                                                                param_init_type=param_init_type)
        adapter_dict["mindpet_delta_adapter_non_linear"] = get_activation(non_linearity)
        adapter_dict["mindpet_delta_adapter_up_sampler"] = _Linear(bottleneck_size,
                                                              hidden_size,
                                                              compute_dtype=compute_dtype,
                                                              param_init_type=param_init_type)

        self.mindpet_delta_adapter_block = nn.SequentialCell(adapter_dict)
        self.residual_add = P.Add()
        self.cast = P.Cast()
        self.shape = P.Shape()
        self.reshape = P.Reshape()

    def construct(self, input_tensor):
        """Foward"""
        # get input_tensor info
        input_tensor_shape = self.shape(input_tensor)
        ori_dtype = F.dtype(input_tensor)

        # reshape input_tensor to compute
        input_tensor = self.reshape(input_tensor, (-1, input_tensor_shape[-1]))

        # calculate adapter_out
        adapter_out = self.mindpet_delta_adapter_block(input_tensor)

        # residual connection, add input and adapter_out
        output = self.residual_add(input_tensor, adapter_out)

        # recover the previous outshape and dtype
        out_shape = input_tensor_shape[:-1] + (-1, )
        output = self.reshape(output, out_shape)
        output = self.cast(output, ori_dtype)
        return output

    def shard(self,
              strategy_matmul_down_sampler=None,
              strategy_bias_down_sampler=None,
              strategy_non_linearity=None,
              strategy_matmul_up_sampler=None,
              strategy_bias_up_sampler=None,
              strategy_residual_add=None):
        """Shard Method"""
        try:
            self.mindpet_delta_adapter_block.mindpet_delta_adapter_down_sampler.shard(
                strategy_matmul=strategy_matmul_down_sampler, strategy_bias=strategy_bias_down_sampler)

            if self.non_linearity_name.lower() == "leakyrelu":
                self.mindpet_delta_adapter_block.mindpet_delta_adapter_non_linear.select_op.shard(
                    (strategy_non_linearity[0], strategy_non_linearity[0]))
            elif self.non_linearity_name.lower() == "logsigmoid":
                self.mindpet_delta_adapter_block.mindpet_delta_adapter_non_linear.mul.shard((
                    strategy_non_linearity[0], ()))
                self.mindpet_delta_adapter_block.mindpet_delta_adapter_non_linear.exp.shard(strategy_non_linearity)
                self.mindpet_delta_adapter_block.mindpet_delta_adapter_non_linear.add.shard((
                    strategy_non_linearity[0], ()))
                self.mindpet_delta_adapter_block.mindpet_delta_adapter_non_linear.rec.shard(strategy_non_linearity)
                self.mindpet_delta_adapter_block.mindpet_delta_adapter_non_linear.log.shard(strategy_non_linearity)
            elif self.non_linearity_name.lower() == "logsoftmax":
                raise ValueError("The 'LogSoftmax' function is not supported in semi auto parallel "
                                "or auto parallel mode.")
            else:
                getattr(self.mindpet_delta_adapter_block.mindpet_delta_adapter_non_linear,
                    self.non_linearity_name).shard(strategy_non_linearity)

            self.mindpet_delta_adapter_block.mindpet_delta_adapter_up_sampler.shard(
                strategy_matmul=strategy_matmul_up_sampler,strategy_bias=strategy_bias_up_sampler)

            self.residual_add.shard(strategy_residual_add)

        except Exception as ex:
            # pylint: disable=W0719
            raise Exception(f"Exception occurred when set the shard for AdapterLayer, error message: \
                {str(ex)}") from ex


class AdapterDense(nn.Dense):
    """
    定义微调算法adapter dense层，继承nn.Dense。

    Args:
        in_channels (int): adapter dense层输入Tensor的空间维度。
        out_channels (int): adapter dense层输出Tensor的空间维度。
        weight_init (Union[Tensor, str, Initializer, numbers.Number]):
            线性层权重参数的初始化方法。
            它的类型可以是Tensor，str，Initializer或numbers.Number。
            当使用str时，值引用自类initializer；更多细节请参考Initializer的值。
            当使用Tensor时，数据类型与输入Tensor相同。
            默认值："normal"。
        bias_init (Union[Tensor, str, Initializer, numbers.Number]):
            线性层偏置参数的初始化方法。
            它的类型可以是Tensor，str，Initializer或numbers.Number。
            当使用str时，值引用自类initializer；更多细节请参考Initializer的值。
            当使用Tensor时，数据类型与输入Tensor相同。
            默认值："zeros"。
        has_bias (int): 是否有偏置。
        activation (Union[str, Cell, Primitive, None]): 激活函数，是与创建层的输入具有相同数据类型的权重矩阵。
        bottleneck_size (int): adapter模块中bottleneck的隐藏大小，默认值：64。
        non_linearity (str): adapter模块中bottleneck非线性激活函数，
            str的值引用自mindspore中 `get_activation` 方法所支持的激活函数类型.默认值：'gelu'。
        param_init_type (dtype.Number): 表示dense中模块的参数初始化类型。
            其值应为dtype.float32或dtype.float16。
            默认值：dtype.float32。
        compute_dtype (dtype.Number): 表示dense中矩阵乘法的计算类型。
            其值应为dtype.float32或dtype.float16。
            默认值：dtype.float16。

    Inputs:
        x (Tensor): 网络的所有输入组成的元组，shape为(*, in_channels)的Tensor，
            in_channels与入参中的in_channels类型一致。
    Outputs:
        output (Tensor): adapter dense的计算结果，shape为 (*, out_channels)，
            out_channels与入参中的out_channels类型一致。
    """

    @_args_type_validator_check(in_channels=Validator.check_positive_int,
                                out_channels=Validator.check_positive_int,
                                has_bias=Validator.check_bool,
                                bottleneck_size=Validator.check_positive_int,
                                non_linearity=_valid_value_checks(list(_activation.keys()), "AdapterDense"),
                                param_init_type=_valid_value_checks(VALID_TENSOR_DATATYPE, "AdapterDense"),
                                compute_dtype=_valid_value_checks(VALID_TENSOR_DATATYPE, "AdapterDense"))

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 weight_init="normal",
                 bias_init="zeros",
                 has_bias: bool = True,
                 activation=None,
                 bottleneck_size: int = 64,
                 non_linearity: str = "gelu",
                 param_init_type: mstype = mstype.float32,
                 compute_dtype: mstype = mstype.float16):
        super().__init__(in_channels=in_channels,
                         out_channels=out_channels,
                         weight_init=weight_init,
                         bias_init=bias_init,
                         has_bias=has_bias,
                         activation=activation)

        self.mindpet_delta_adapter = AdapterLayer(hidden_size=out_channels,
                                             bottleneck_size=bottleneck_size,
                                             non_linearity=non_linearity,
                                             param_init_type=param_init_type,
                                             compute_dtype=compute_dtype)

        self.bottleneck_size = bottleneck_size
        self.compt_dtype = compute_dtype
        self.cast = P.Cast()
        self.act_name = activation

    def construct(self, input_tensor):
        """Foward"""
        # get input_x info
        x_shape = self.shape_op(input_tensor)
        ori_dtype = F.dtype(input_tensor)

        # reshape input_tensor to compute
        input_tensor = self.reshape(input_tensor, (-1, x_shape[-1]))

        # start to linear compute
        weight = self.cast(self.weight, self.compt_dtype)
        input_tensor = self.cast(input_tensor, self.compt_dtype)

        input_tensor = self.matmul(input_tensor, weight)
        if self.has_bias:
            input_tensor = self.bias_add(
                input_tensor, self.cast(self.bias, self.compt_dtype))
        if self.activation_flag:
            input_tensor = self.activation(input_tensor)

        # calculate adapter_out
        input_tensor = self.mindpet_delta_adapter(input_tensor)

        # recover the previous outshape and dtype
        out_shape = x_shape[:-1] + (-1,)
        input_tensor = self.reshape(input_tensor, out_shape)
        output = self.cast(input_tensor, ori_dtype)
        return output

    def shard(self,
              strategy_matmul_org=None,
              strategy_bias_org=None,
              strategy_activation_org=None,
              strategy_matmul_down_sampler=None,
              strategy_bias_down_sampler=None,
              strategy_non_linearity=None,
              strategy_matmul_up_sampler=None,
              strategy_bias_up_sampler=None,
              strategy_residual_add=None):
        """Shard Method"""
        try:
            # set origin dense strategy
            self.matmul.shard(strategy_matmul_org)
            if self.has_bias:
                self.bias_add.shard(strategy_bias_org)
            if self.activation_flag and isinstance(self.act_name, str):
                if self.act_name.lower() == "leakyrelu":
                    self.activation.select_op.shard(
                        (strategy_activation_org[0], strategy_activation_org[0]))
                elif self.act_name.lower() == "logsigmoid":
                    self.activation.mul.shard((strategy_activation_org[0], ()))
                    self.activation.exp.shard(strategy_activation_org)
                    self.activation.add.shard((strategy_activation_org[0], ()))
                    self.activation.rec.shard(strategy_activation_org)
                    self.activation.log.shard(strategy_activation_org)
                elif self.act_name.lower() == "logsoftmax":
                    raise ValueError("The 'LogSoftmax' function is not supported in semi auto parallel "
                                    "or auto parallel mode.")
                else:
                    getattr(self.activation, self.act_name).shard(strategy_activation_org)

            # set adapter strategy
            self.mindpet_delta_adapter.shard(strategy_matmul_down_sampler=strategy_matmul_down_sampler,
                                        strategy_bias_down_sampler=strategy_bias_down_sampler,
                                        strategy_non_linearity=strategy_non_linearity,
                                        strategy_matmul_up_sampler=strategy_matmul_up_sampler,
                                        strategy_bias_up_sampler=strategy_bias_up_sampler,
                                        strategy_residual_add=strategy_residual_add)

        except Exception as ex:
            # pylint: disable=W0719
            raise Exception(f"Exception occurred when set the shard for AdapterDense, error message: \
                {str(ex)}") from ex



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved."""

from mindpet.task.options import DataPathOption, PretrainedModelPathOption, OutputPathOption, BootFilePathOption, \
    ModelConfigPathOption, QuietOption, CkptPathOption, TimeoutOption


def finetune_options():
    """
    微调功能入参集合
    :return: 装饰器
    """
    options = [
        QuietOption(),
        DataPathOption(),
        PretrainedModelPathOption(),
        OutputPathOption(),
        BootFilePathOption(),
        ModelConfigPathOption(),
        TimeoutOption()
    ]

    def decorator(func):
        if not hasattr(func, '__click_params__'):
            func.__click_params__ = []
        func.__click_params__ += options
        return func

    return decorator


def evaluate_options():
    """
    评估功能入参集合
    :return: 装饰器
    """
    options = [
        QuietOption(),
        DataPathOption(),
        CkptPathOption(),
        OutputPathOption(),
        BootFilePathOption(),
        ModelConfigPathOption(),
        TimeoutOption()
    ]

    def decorator(func):
        if not hasattr(func, '__click_params__'):
            func.__click_params__ = []
        func.__click_params__ += options
        return func

    return decorator


def infer_options():
    """
    推理功能入参集合
    :return: 装饰器
    """
    options = [
        QuietOption(),
        DataPathOption(),
        CkptPathOption(),
        OutputPathOption(),
        BootFilePathOption(),
        ModelConfigPathOption(),
        TimeoutOption()
    ]

    def decorator(func):
        if not hasattr(func, '__click_params__'):
            func.__click_params__ = []
        func.__click_params__ += options
        return func

    return decorator



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""Model Config check module."""

from mindpet.log.log import logger
from mindpet.utils.exceptions import ModelConfigParamsInfoError
from mindpet.utils.constants import MODEL_CONFIG_PARAMS_CHARACTER_BLACKLIST, MODEL_CONFIG_LEN_LIMIT, \
    MINDPET_DEFINED_PARAM_NAME, INVALID_CUSTOM_PARAM_KEY_PREFIX, INVALID_CUSTOM_PARAM_VAL_PREFIX


class ModelConfigParamsChecker:
    """
    Define model config params checker.
    """
    def __init__(self, task_object, params_config=None):
        """
        model_config配置文件内容校验构造方法
        :param task_object: 任务对象
        :param params_config: params部分参数
        """
        if task_object is None:
            raise ValueError('Param [task_object] is required.')

        if not params_config:
            raise ValueError('Param [params_config] is required.')

        if params_config is not None and not isinstance(params_config, dict):
            raise TypeError('Invalid type for param [params_config], which must be key-value pairs.')

        self.task_object = task_object
        self.params = params_config

    def check(self):
        """
        model_config配置文件params部分参数校验
        """
        for param in self.params.items():
            param_key, param_val = param[0], param[1]

            if param_key is None or param_val is None:
                raise ModelConfigParamsInfoError('Config in [params] part in model config file is empty.')

            # params的value部分是否为单层键值对
            if isinstance(param_val, dict):
                raise ModelConfigParamsInfoError(
                    'Config in [params] part in model config file only supports single-level nesting.')

            param_key, param_val = str(param_key), str(param_val)

            # 接口预定义参数与params的key重复校验
            if param_key in MINDPET_DEFINED_PARAM_NAME:
                # pylint: disable=W1203
                logger.warning(
                    f'Find duplicate key [{param_key}] from config in [params] part in model config file.')
                continue

            # params键值对长度及字符合法性
            if len(param_key) > MODEL_CONFIG_LEN_LIMIT or len(param_val) > MODEL_CONFIG_LEN_LIMIT:
                raise ModelConfigParamsInfoError('Config in [params] part in model config file is too long.')

            for k_item in param_key:
                if k_item in MODEL_CONFIG_PARAMS_CHARACTER_BLACKLIST:
                    raise ModelConfigParamsInfoError('Config in [params] part in model config file '
                                                     'contains invalid character(s).')

            for v_item in param_val:
                if v_item in MODEL_CONFIG_PARAMS_CHARACTER_BLACKLIST:
                    raise ModelConfigParamsInfoError('Config in [params] part in model config file '
                                                     'contains invalid character(s).')

            # params键值对不合法前缀校验
            if param_key.startswith(INVALID_CUSTOM_PARAM_KEY_PREFIX) or param_val.startswith(
                    INVALID_CUSTOM_PARAM_VAL_PREFIX):
                raise ModelConfigParamsInfoError('Invalid character(s) in [params] part in model config file.')

            self.task_object.command_params.extend([f'--{param_key}', param_val])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""base check module."""
from mindpet.security.param_check.option_check_utils import PathRightEscalationCheck, FileSizeCheck


class BaseCheckParam:
    """base param check class."""
    def __init__(self, mode, path_including_file, force_quit, quiet):
        """
        入参校验参数
        :param mode: 路径权限范围约束
        :param path_including_file: 路径是否为文件路径
        :param force_quit: SDK场景下, 路径权限提升校验异常时是否强制退出
        :param quiet: CLI场景下，路径权限提升校验是否忽略交互式过程
        """
        self.mode = mode
        self.path_including_file = path_including_file
        self.force_quit = force_quit
        self.quiet = quiet


class BaseCheck:
    """base check class."""
    def __init__(self, option_name, option_value):
        """
        基础校验项构造方法
        :param option_name: 入参名称
        :param option_value: 入参值
        """
        self.option_name = option_name
        self.option_value = option_value

    def check(self, check_param):
        """
        微调入参校验
        :param check_param: 微调入参校验参数
        """
        self.base_check_item(check_param)

    def base_check_item(self, check_param):
        """
        基础校验项
        :param check_param: 校验参数
        """
        # 路径权限提升校验
        PathRightEscalationCheck(option_name=self.option_name,
                                 option_value=self.option_value,
                                 mode=check_param.mode,
                                 force_quit=check_param.force_quit,
                                 quiet=check_param.quiet)

        # 路径文件大小校验
        FileSizeCheck(option_name=self.option_name,
                      option_value=self.option_value,
                      path_including_file=check_param.path_including_file)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved."""

from mindpet.security.param_check.base_check import BaseCheckParam, BaseCheck
from mindpet.security.param_check.option_check_utils import PathExistCheck, LinkPathCheck, get_real_path


class ResultFileCheckParam(BaseCheckParam):
    """ResultFileCheckParam class"""
    def __init__(self,
                 path_content_check_param,
                 mode,
                 path_including_file,
                 force_quit,
                 quiet):
        """
        微调入参校验参数
        :param path_content_check_param: 路径内容校验参数
        :param mode: 路径权限范围约束
        :param path_including_file: 路径是否为文件路径
        :param force_quit: SDK场景下, 路径权限提升校验异常时是否强制退出
        """
        super().__init__(mode, path_including_file, force_quit, quiet)
        self.path_content_check_param = path_content_check_param


# pylint: disable=W0246
class ResultFileCheck(BaseCheck):
    """ResultFileCheck class"""
    def check(self, check_param):
        """
        任务结果文件(eval_result.json/infer_result.json)合法性校验
        :param check_param: 任务结果文件校验参数
        """
        # 针对原始路径进行的校验
        self._origin_path_check_item()

        # 获取真实路径
        self.option_value = get_real_path(self.option_value)

        # 针对真实路径进行的校验
        self.base_check_item(check_param)

    def _origin_path_check_item(self):
        # 路径真实性校验
        PathExistCheck(option_name=self.option_name, option_value=self.option_value)

        # 路径软链接校验
        LinkPathCheck(option_name=self.option_name, option_value=self.option_value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""

from mindpet.log.log import get_logger, logger, operation_logger_without_std, operation_logger, logger_without_std

__all__ = ['get_logger', 'logger', 'logger_without_std', 'operation_logger', 'operation_logger_without_std']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""Task APIs"""
import os
import signal
import time
import uuid
import multiprocessing
from pathlib import Path
import click
from mindpet.security.param_check.model_config_params_check_util import ModelConfigParamsChecker
from mindpet.utils.exceptions import MakeDirError, PathLoopError, ModelConfigKeysInfoError, ModelConfigParamsInfoError
from mindpet.log.log import logger, logger_without_std, record_operation_and_service_info_log
from mindpet.utils.constants import OUTPUT_PATH_RANDOM_DIR_MODE, TIME_MONITOR_SLEEP_TIME, SECONDS_PER_HOUR, HOURS_PER_DAY


def monitor_process_rsp_code(task_name, task_process, timeout=None):
    """
    含超时机制的子进程监测器, 并获取子进程结束状态码
    :param task_name: 任务名称
    :param task_process: 待监听进程
    :param timeout: 超时时间(秒数)
    """
    timeout_monitor_process = None

    # 启动异步进程做超时监听
    if timeout is not None:
        timeout_monitor_process = multiprocessing.Process(target=timeout_monitor,
                                                          args=(task_name, task_process, timeout))
        timeout_monitor_process.start()

    # 持续监听子进程标准输出管道, 捕捉管道内容并落盘
    while task_process.poll() is None:
        process_log = str(task_process.stdout.readline().decode('utf-8')).rstrip()
        if process_log:
            logger.info(process_log, extra={'Model': True})

    # 子进程正常退出时, 如超时监听进程未结束, 结束超时监听
    if timeout is not None and timeout_monitor_process is not None and timeout_monitor_process.is_alive():
        timeout_monitor_process.terminate()

    return task_process.poll()


def timeout_monitor(task_name, task_process, timeout):
    """
    超时监听, 超时后终止被监听进程
    :param task_name: 任务名称
    :param task_process: 被监听进程
    :param timeout: 超时时间(秒数)
    """
    begin_time = time.time()
    while True:
        if time.time() - begin_time > timeout + 1:
            days = int(timeout // SECONDS_PER_HOUR // HOURS_PER_DAY)
            hours = int((timeout // SECONDS_PER_HOUR) % HOURS_PER_DAY)

            logger.warning('Task exceeds the time limit, it is forcibly terminated [%ddays %dhours].', days, hours)
            record_operation_and_service_info_log('%s failed because of time limit.', str(task_name).capitalize())
            task_process.kill()
            task_process.terminate()

            # 避免子进程无法全部结束, 向同进程组进程发送结束信号
            try:
                os.killpg(os.getpgid(os.getpid()), signal.SIGTERM)
            except (TypeError, ProcessLookupError):
                logger_without_std.warning('Task exceeds the time limit, it is forcibly terminated [%ddays %dhours].',
                                           days, hours)
                break

        time.sleep(TIME_MONITOR_SLEEP_TIME)


def create_output_path_subdir_with_uuid(output_path):
    """
    在output_path下生成包含uuid随机片段的子文件夹名称
    :param output_path: 入参output_path值
    """
    uuid_name = str(uuid.uuid4()).replace('-', '')
    random_dir = f'mindpet_{uuid_name}'.upper()
    full_output_path = os.path.join(output_path, random_dir).replace('\\', '/')

    while os.path.exists(full_output_path):
        uuid_name = str(uuid.uuid4()).replace('-', '')
        random_dir = f'mindpet_{uuid_name}'.upper()
        full_output_path = os.path.join(output_path, random_dir).replace('\\', '/')

    try:
        os.makedirs(full_output_path, exist_ok=True, mode=OUTPUT_PATH_RANDOM_DIR_MODE)
    except Exception as ex:
        raise MakeDirError(f'Failed to create directory from param [output_path], error message: {str(ex)}') from ex

    # pylint: disable=W1203
    logger.info(f'Create output directory successfully, directory name: {random_dir}')
    return full_output_path


def model_config_keys_check_item(content, config_keys):
    """
    校验model_config.yaml文件中的配置头名称是否合法
    :param content: yaml文件中配置内容
    :param config_keys: 允许配置的头名称
    """
    if not isinstance(config_keys, tuple):
        config_keys = (config_keys,)

    try:
        for key_item in content.keys():
            if key_item not in config_keys:
                raise ModelConfigKeysInfoError("Invalid config in model config file, "
                                               f"only support [{'/'.join(config_keys)}]")
    except AttributeError as ex:
        raise ModelConfigKeysInfoError('Invalid key in model config file.') from ex


def model_config_params_check_item(task_object, content):
    """
    model_config自定义参数内容黑名单校验及参数拼装
    :param task_object: 任务对象
    :param content: 待校验内容
    """
    if 'params' in content.keys():
        params_config = content.get('params')

        if params_config is None:
            raise ModelConfigParamsInfoError('[params] attribute in model config file is empty.')

        params_check = ModelConfigParamsChecker(task_object=task_object, params_config=params_config)
        params_check.check()


def extend_model_config_freeze_command(task_object, model_config_path, content):
    """
    如果有freeze关键词，则拼接--advanced_config参数传递文件路径
    :param content: 待校验内容
    :param task_object: 参数对象
    :param model_config_path: model_config配置文件路径
    """
    if 'freeze' in content.keys():
        task_object.command_params.extend(['--advanced_config', model_config_path])


def path_in_out_loop_check(param_dict):
    """
    用于判断输入、输出路径是否存在回环可能
    :param param_dict: 输入参数集合
    """
    # 从参数集合中提取输出路径
    output_path = Path(param_dict.get('output_path'))

    # 从参数集合中提取输入路径
    input_path_dict = {'data_path': Path(param_dict.get('data_path'))}
    if param_dict.get('pretrained_model_path') is not None:
        input_path_dict['pretrained_model_path'] = Path(param_dict.get('pretrained_model_path'))
    if param_dict.get('ckpt_path') is not None:
        input_path_dict['ckpt_path'] = Path(param_dict.get('ckpt_path'))

    # 判断输入路径/输出径是否相等或输出路径是否是输入路径的子路径, 如果是, 则存在回环路径的可能
    for input_path_name, input_path_value in input_path_dict.items():
        if (output_path == input_path_value) or (input_path_value in output_path.parents):
            raise PathLoopError(f'Param [output_path] can not be the same as, '
                                f'or a subdirectory of param [{input_path_name}]')


def handle_exception_log(exception):
    """
    根据异常场景, 给出对应异常日志
    :param exception: 异常场景
    """
    if isinstance(exception, click.exceptions.Abort):
        logger.info('Current command is artificially canceled.')
    elif isinstance(exception, click.exceptions.NoSuchOption):
        # pylint: disable=W1203
        logger.error(f'Invalid param detected, error message: {str(exception)}')
    elif isinstance(exception, click.exceptions.MissingParameter):
        # pylint: disable=W1203
        logger.error(f'Necessary param is missing, error message: {str(exception)}')
    elif isinstance(exception, Exception):
        if exception is None or not str(exception):
            logger.error('Exception occurred, no error message available.')
        elif str(exception).isdigit():
            # pylint: disable=W1203
            logger.error(f'Exception occurred, error code: {str(exception)}')
        # pylint: disable=W1203
        logger.error(f'Exception occurred, error message: {str(exception)}')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""command option check utils."""

import os
from mindpet.log.log import logger
from mindpet.utils.io_utils import is_link_path
from mindpet.utils.entrance_monitor import entrance_monitor
from mindpet.utils.constants import EMPTY_STRING, BLACKLIST_CHARACTERS, BASE_WHITELIST_CHARACTERS, MAX_FILE_MB_SIZE, \
    MB_SIZE, ENTRANCE_TYPE, MIN_MODE, MAX_MODE, GB_SIZE
from mindpet.utils.exceptions import FileOversizeError, LinkPathError, LowDiskFreeSizeRiskError, PathContentError, \
    AbsolutePathError, PathLengthError, FileNameLengthError, PathGranularityError, PathOwnerError, PathModeError, \
    PathRightEscalationError, ManualCancelError


def get_real_path(path):
    """
    获取给定路径对应的真实路径

    :param path: 给定路径
    :return: 给定路径对应真实路径
    """
    if path is None or path == EMPTY_STRING:
        raise ValueError('Path is None or empty when getting real path.')

    return os.path.realpath(path)


class OptionBase:
    """Define command option."""
    def __init__(self, option_name, option_value):
        """
        构造方法
        :param option_name: 入参名称
        :param option_value: 入参值
        """
        if option_name is None or option_name == EMPTY_STRING:
            raise ValueError('Param [option_name] is None or empty.')

        if option_value is None or option_value == EMPTY_STRING:
            raise ValueError('Param [option_value] is None or empty.')

        self.option_name = option_name
        self.option_value = option_value

    def _is_file_path(self):
        """
        判断路径是否是一个文件
        :return: 是否是一个文件
        """
        return os.path.isfile(self.option_value)

    def _is_dir_path(self):
        """
        判断路径是否是一个文件夹
        :return: 是否是一个文件夹
        """
        return os.path.isdir(self.option_value)


class PathLengthCheckParam:
    """Define file path length check params."""
    def __init__(self, path_min_limit, path_max_limit, file_min_limit, file_max_limit):
        """
        路径长度校验项参数构造方法
        :param path_min_limit: 路径最小长度约束
        :param path_max_limit: 路径最大长度约束
        :param file_min_limit: 文件名最小长度约束
        :param file_max_limit: 文件名最大长度约束
        """
        self.path_min_limit = path_min_limit
        self.path_max_limit = path_max_limit
        self.file_min_limit = file_min_limit
        self.file_max_limit = file_max_limit


class PathContentCheckParam:
    """Define file path check params."""
    def __init__(self, base_whitelist_mode, extra_whitelist):
        """
        路径内容校验项参数构造方法
        :param base_whitelist_mode: 路径基础白名单模式
        :param extra_whitelist: 路径额外字符白名单
        """
        self.base_whitelist_mode = base_whitelist_mode
        self.extra_whitelist = extra_whitelist


class InteractionByEntrance(OptionBase):
    """Define interaction."""
    def __init__(self, option_name, option_value, notice_msg, notice_accept_msg, exception_type):
        """
        根据使用入口(CLI/SDK)执行不同的交互逻辑构造方法
        :param option_name: 参数名称
        :param option_value: 参数值
        :param notice_msg: 通知消息
        :param notice_accept_msg: 用户接受通知消息
        :param exception_type: 异常类型
        """
        super().__init__(option_name, option_value)

        if notice_msg is None or notice_msg == EMPTY_STRING:
            raise ValueError('Param [notice_msg] is None or empty.')

        if not isinstance(notice_msg, str):
            raise TypeError('Invalid type for param [notice_msg] when initializing InteractionByEntrance.')

        if notice_accept_msg is None or notice_accept_msg == EMPTY_STRING:
            raise ValueError('Param [notice_accept_msg] is None or empty.')

        if not isinstance(notice_accept_msg, str):
            raise TypeError('Invalid type for param [notice_accept_msg] when initializing InteractionByEntrance.')

        if exception_type is None:
            raise ValueError('Param [exception_type] is None.')

        if not issubclass(exception_type, Exception):
            raise TypeError('Invalid type for param [exception_type] when initializing InteractionByEntrance.')

        self.notice_msg = notice_msg
        self.notice_accept_msg = notice_accept_msg
        self.exception_type = exception_type

    def interact_by_entrance(self, force_quit, quiet):
        """
        根据不同的入口(CLI/SDK), 进行不同类型权限校验交互
        :param force_quit: 触发校验异常时，是否强制结束进程
        :param quiet: 是否启用安静模式
        """
        if force_quit is None:
            raise ValueError('Param [force_quit] is None.')

        if quiet is None:
            raise ValueError('Param [quiet] is None.')

        if not isinstance(force_quit, bool):
            raise TypeError('Invalid type for param [force_quit] in interact_by_entrance.')

        if not isinstance(quiet, bool):
            raise TypeError('Invalid type for param [quiet] in interact_by_entrance.')

        entrance_type = entrance_monitor.get_value(ENTRANCE_TYPE)

        if entrance_type == 'CLI':
            self._interact_by_cli(force_quit, quiet)
        elif entrance_type == 'SDK':
            self._interact_by_sdk(force_quit)
        else:
            raise ValueError('Invalid param [entrance_type].')

    def _interact_by_cli(self, force_quit, quiet):
        """
        CLI场景下的交互方式
        :param force_quit: 是否强制终止请求
        :param quiet: 是否启动安静模式
        """
        if quiet:
            if not force_quit:
                logger.warning(self.notice_msg)
            else:
                raise self.exception_type(self.notice_msg)
        else:
            ans = input(self.notice_msg + ' would you like to continue? (y/n)')
            if str(ans).lower() != 'y':
                raise ManualCancelError('Current command is cancelled.')
            logger.info(self.notice_accept_msg)

    def _interact_by_sdk(self, force_quit):
        """
        SDK场景下的交互方式
        :param force_quit: 是否强制终止请求
        """
        if force_quit:
            raise self.exception_type(self.notice_msg)
        logger.warning(self.notice_msg)


class PathContentBlacklistCharactersCheck(OptionBase):
    """Define path black list check"""
    def __init__(self, option_name, option_value):
        """
        路径内容黑名单字符校验构造方法
        :param option_name: 参数名称
        :param option_value: 参数值
        """
        super().__init__(option_name, option_value)
        self._check()

    def _check(self):
        """
        路径内容黑名单字符校验
        """
        self.option_value = str(self.option_value).strip()

        for item in BLACKLIST_CHARACTERS:
            if item in self.option_value:
                raise PathContentError(f'Invalid character(s) in param [{self.option_name}].')


class AbsolutePathCheck(OptionBase):
    """Define absolute path check."""
    def __init__(self, option_name, option_value):
        """
        绝对路径校验构造方法
        :param option_name: 参数名称
        :param option_value: 参数值
        """
        super().__init__(option_name, option_value)
        self._check()

    def _check(self):
        """
        绝对路径校验
        """
        if not os.path.isabs(self.option_value):
            raise AbsolutePathError(f'Param [{self.option_name}] is not an absolute path.')


class PathExistCheck(OptionBase):
    """Define path exist check."""
    def __init__(self, option_name, option_value):
        """
        路径存在校验构造方法
        :param option_name: 参数名称
        :param option_value: 参数值
        """
        super().__init__(option_name, option_value)
        self._check()

    def _check(self):
        """
        路径存在校验
        """
        if not os.path.exists(self.option_value):
            raise FileNotFoundError(f'Path of param [{self.option_name}] does not exist.')


class LinkPathCheck(OptionBase):
    """Define link path exist check."""
    def __init__(self, option_name, option_value):
        """
        链接路径校验构造方法
        :param option_name: 参数名称
        :param option_value: 参数值
        """
        super().__init__(option_name, option_value)
        self._check()

    def _check(self):
        """
        链接路径校验
        """
        if is_link_path(self.option_value):
            raise LinkPathError(f'Path of param [{self.option_name}] is a link path.')


class PathContentLengthCheck(OptionBase):
    """Define path content option length check."""
    def __init__(self, option_name, option_value, path_min_limit, path_max_limit, file_min_limit, file_max_limit):
        """
        路径内容长度校验构造方法
        :param option_name: 参数值
        :param option_value: 参数方法
        :param path_min_limit: 路径最小长度限制
        :param path_max_limit: 路径最大长度限制
        :param file_min_limit: 路径中文件名最小长度限制
        :param file_max_limit: 路径中文件名最大长度限制
        """
        if path_min_limit is not None and not isinstance(path_min_limit, int):
            raise TypeError('Invalid type for param [path_min_limit] when initializing PathContentLengthCheck.')

        if path_max_limit is not None and not isinstance(path_max_limit, int):
            raise TypeError('Invalid type for param [path_max_limit] when initializing PathContentLengthCheck.')

        if file_min_limit is not None and not isinstance(file_min_limit, int):
            raise TypeError('Invalid type for param [file_min_limit] when initializing PathContentLengthCheck.')

        if file_max_limit is not None and not isinstance(file_max_limit, int):
            raise TypeError('Invalid type for param [file_max_limit] when initializing PathContentLengthCheck.')

        super().__init__(option_name, option_value)
        self._check(path_min_limit, path_max_limit, file_min_limit, file_max_limit)

    def _check(self, path_min_limit, path_max_limit, file_min_limit, file_max_limit):
        """
        路径内容长度校验
        :param path_min_limit: 路径最小长度, 默认为0
        :param path_max_limit: 路径最大长度
        :param file_min_limit: 文件最小长度
        :param file_max_limit: 文件最大长度
        """
        if not self._path_length_check_item(path_min_limit=path_min_limit, path_max_limit=path_max_limit):
            raise PathLengthError(f'Length of path in param [{self.option_name}] is invalid.')

        if self._is_file_path():
            if not self._file_length_check_item(file_min_limit=file_min_limit, file_max_limit=file_max_limit):
                raise FileNameLengthError(f'Length of file name in param [{self.option_name}] is invalid.')

    def _path_length_check_item(self, path_min_limit, path_max_limit):
        """
        路径长度校验
        :param path_min_limit: 路径最小长度限制, 默认值为0
        :param path_max_limit: 路径最大长度限制
        :return: 路径粒度长度是否合规
        """
        path_min_limit = 0 if path_min_limit is None else path_min_limit

        if path_min_limit is not None and path_min_limit < 0:
            logger.error('Param [path_min_limit] should be non-negative integer.')
            return False

        if path_max_limit is not None and path_max_limit < 0:
            logger.error('Param [path_max_limit] should be non-negative integer.')
            return False

        if path_min_limit is not None and path_max_limit is not None and path_min_limit > path_max_limit:
            logger.error('Param [path_min_limit] should be less than or equal to path_max_limit.')
            return False

        if path_min_limit is not None and len(self.option_value) < path_min_limit:
            return False

        if path_max_limit is not None and len(self.option_value) > path_max_limit:
            return False

        return True

    def _file_length_check_item(self, file_min_limit, file_max_limit):
        """
        路径中文件名长度校验
        :param file_min_limit: 文件名最小长度限制, 默认值为0
        :param file_max_limit: 文件名最大长度限制
        :return: 文件名粒度长度是否合规
        """
        file_name = os.path.basename(self.option_value)
        file_min_limit = 0 if file_min_limit is None else file_min_limit

        if file_min_limit is not None and file_min_limit < 0:
            logger.error('Param [file_min_limit] should be non-negative integer.')
            return False

        if file_max_limit is not None and file_max_limit < 0:
            logger.error('Param [file_max_limit] should be non-negative integer.')
            return False

        if file_min_limit is not None and file_max_limit is not None and file_min_limit > file_max_limit:
            logger.error('Param [file_min_limit] should be less than or equal to file_max_limit.')
            return False

        if file_min_limit is not None and len(file_name) < file_min_limit:
            return False

        if file_max_limit is not None and len(file_name) > file_max_limit:
            return False

        return True


class PathContentCharacterCheck(OptionBase):
    """Define path content option character check."""
    def __init__(self, option_name, option_value, base_whitelist_mode, extra_whitelist):
        """
        路径内容校验构造方法
        :param option_name: 参数名称
        :param option_value: 参数值
        :param base_whitelist_mode: 基础白名单模式
        :param extra_whitelist: 额外白名单列表
        """
        super().__init__(option_name, option_value)

        if base_whitelist_mode is None or base_whitelist_mode == EMPTY_STRING:
            raise ValueError('Param [base_whitelist_mode] is None or empty.')

        if not isinstance(base_whitelist_mode, str):
            raise TypeError('Invalid type for param [base_whitelist_mode] when initializing PathContentCharacterCheck.')

        if extra_whitelist is not None and not isinstance(extra_whitelist, list):
            raise TypeError('Invalid type for param [extra_whitelist] when initializing PathContentCharacterCheck.')

        self._check(base_whitelist_mode, extra_whitelist)

    @staticmethod
    def get_base_whitelist_by_mode(base_whitelist_mode):
        """
        根据base_whitelist_mode获得对应字符粒度参数校验基础白名单
        :param base_whitelist_mode: 模式类型，包括全大写/全小写/数字/大小写字母/大小写字母和数字
        :return: 基础白名单字符列表
        """
        if base_whitelist_mode not in BASE_WHITELIST_CHARACTERS:
            raise ValueError('Invalid param [base_whitelist_mode], only support %s' %
                             list(BASE_WHITELIST_CHARACTERS.keys()))

        return BASE_WHITELIST_CHARACTERS.get(base_whitelist_mode)

    def _check(self, base_whitelist_mode, extra_whitelist):
        """
        路径内容校验
        :param base_whitelist_mode: 基础白名单模式
        :param extra_whitelist: 额外补充白名单字符
        """
        base_whitelist_mode = self.get_base_whitelist_by_mode(base_whitelist_mode)
        extra_whitelist = [] if extra_whitelist is None else extra_whitelist
        whitelist = base_whitelist_mode + extra_whitelist

        for val in self.option_value:
            if val not in whitelist:
                raise PathContentError(f'Invalid character(s) in param [{self.option_name}].')


class PathGranularityCheck(OptionBase):
    """Define path option granular check."""
    def __init__(self, option_name, option_value, path_including_file):
        """
        路径粒度校验(文件夹/文件粒度)构造方法
        :param option_name: 参数名称
        :param option_value: 参数值
        :param path_including_file: 路径是否包含文件名, 传None表示同时兼容
        """
        super().__init__(option_name, option_value)

        if path_including_file is not None and not isinstance(path_including_file, bool):
            raise TypeError('Invalid type for param [path_including_file] when initializing PathGranularityCheck.')

        self.path_including_file = path_including_file
        self._check()

    def _check(self):
        """
        路径粒度校验(文件夹/文件粒度)
        """
        status = True

        if self.path_including_file is not None:
            status = self._is_file_path() if self.path_including_file else self._is_dir_path()

        if not status:
            raise PathGranularityError(
                f"Param [{self.option_name}] should be a {'file' if self.path_including_file else 'dir'} path.")


class PathRightEscalationCheck(OptionBase):
    """Define path right option escalation check."""
    def __init__(self, option_name, option_value, mode, force_quit, quiet):
        """
        路径权限提升校验构造方法
        :param option_name: 参数名称
        :param option_value: 参数值
        :param mode: 权限约束
        :param force_quit: 异常是否强制退出
        :param quiet: 安静模式
        """
        super().__init__(option_name, option_value)
        self._check(mode, force_quit, quiet)

    @staticmethod
    def mode_property_check(mode):
        """
        mode属性合法性检查, 返回合法的路径权限
        :param mode: 路径要求的权限范围, 要求必须是字符串类型, 包含3位数字, 数字范围[0,7], 例如'755'
        :return 路径权限值
        """
        if mode is None:
            raise ValueError('Param [mode] is None.')

        if not isinstance(mode, str):
            raise TypeError('Invalid type for param [mode] in mode_property_check.')

        mode = str(mode)
        if len(mode) != 3:
            raise ValueError('Invalid param [mode] for mode_property_check.')

        for pri in mode:
            if not pri.isdigit():
                raise TypeError('Invalid param [mode] for mode_property_check.')

            pri = int(pri)
            if pri < MIN_MODE or pri > MAX_MODE:
                raise ValueError('Invalid param [mode] for mode_property_check.')

        ret = (int(mode[0]), int(mode[1]), int(mode[2]))

        return ret

    def _check(self, mode, force_quit, quiet):
        """
        路径权限提升校验
        :param mode:路径权限约束
        :param force_quit: 发现权限提升问题是否强制退出
        :param quiet: 安静模式
        """
        is_legal_mode = True

        # 路径属主一致性校验
        try:
            self._path_owner_check()
        except PathOwnerError:
            is_legal_mode = False

        # 路径权限范围校验
        try:
            self._path_mode_check(mode)
        except PathModeError:
            is_legal_mode = False

        if not is_legal_mode:
            notice_msg = f'Param [{self.option_name}] may have risk of rights escalation ' \
                         f'(owner of path from param [{self.option_name}] is not current login user, ' \
                         f'or other users have unnecessary permissions of path from param [{self.option_name}]).'

            notice_accept_msg = f'Current login user has accepted the risk of rights escalation of param ' \
                                f'[{self.option_name}].'

            interaction_op = InteractionByEntrance(option_name=self.option_name,
                                                   option_value=self.option_value,
                                                   notice_msg=notice_msg,
                                                   notice_accept_msg=notice_accept_msg,
                                                   exception_type=PathRightEscalationError)
            interaction_op.interact_by_entrance(force_quit=force_quit, quiet=quiet)

    def _path_owner_check(self):
        """
        路径属主一致性校验, 判断路径属主和当前用户是否一致
        """
        path_owner = os.stat(self.option_value).st_uid
        current_login_user = os.geteuid()

        if path_owner != current_login_user:
            raise PathOwnerError(f'Invalid owner of path of param [{self.option_name}].')

    def _path_mode_check(self, mode):
        """
        路径权限校验
        :param mode: 路径要求的权限范围
        """
        required_mode = self.mode_property_check(mode)
        actual_mode = self._get_path_mode()

        mode_size = len(required_mode)
        for idx in range(mode_size):
            if required_mode[idx] | actual_mode[idx] != required_mode[idx]:
                raise PathModeError(f'Invalid mode of path of param [{self.option_name}].')

    def _get_path_mode(self):
        """
        获得当前路径对应的权限, 分用户权限/同组权限/其他权限, 使用十进制表示
        :return: tuple元素, (用户权限, 同组权限, 其他权限)
        """
        mode = oct(os.stat(self.option_value).st_mode)[-3:]
        ret = (int(mode[0]), int(mode[1]), int(mode[2]))
        return ret


class FileSizeCheck(OptionBase):
    """Define file size check."""
    def __init__(self, option_name, option_value, path_including_file):
        """
        文件大小校验构造方法
        :param option_name: 参数名称
        :param option_value: 参数值
        :param path_including_file: 路径是否包含文件名
        """
        super().__init__(option_name, option_value)

        if path_including_file is None:
            raise ValueError('Param [path_including_file] is None.')

        if not isinstance(path_including_file, bool):
            raise TypeError('Invalid type for param [path_including_file] when initializing FileSizeCheck.')

        self._check(path_including_file)

    def _check(self, path_including_file):
        """
        文件大小校验
        """
        if path_including_file:
            size = os.stat(self.option_value).st_size / MB_SIZE
            if size > MAX_FILE_MB_SIZE:
                raise FileOversizeError(f'File of param [{self.option_name}] is too large.')


class DiskFreeSpaceCheck(OptionBase):
    """Define disk free space check."""
    def __init__(self, option_name, option_value, free_space_limit, force_quit, quiet):
        """
        路径所在磁盘剩余空间校验构造方法
        :param option_name: 参数名称
        :param option_value: 参数值
        :param free_space_limit: 剩余最小空间约束
        :param force_quit: 异常是否强制退出
        :param quiet: 安静模式
        """
        super().__init__(option_name, option_value)

        if free_space_limit is None:
            raise ValueError('Param [free_space_limit] is None.')

        if not isinstance(free_space_limit, int):
            raise TypeError('Invalid type for param [free_space_limit] when initializing DiskFreeSpaceCheck.')

        if free_space_limit < 0:
            raise ValueError('Invalid param [free_space_limit] when initializing DiskFreeSpaceCheck.')

        self.free_space_limit = free_space_limit
        self._check(force_quit, quiet)

    def _check(self, force_quit, quiet):
        """
        路径所在磁盘剩余空间校验
        :param force_quit: 是否强制退出
        :param quiet: 是否安静模式
        """
        info = os.statvfs(self.option_value)
        free_size = info.f_bsize * info.f_bavail

        # pylint: disable=W1203
        logger.info(
            f'Disk where param {self.option_name} is located has {round(free_size / GB_SIZE, 2)} GB free space.')

        if free_size <= self.free_space_limit:
            notice_msg = f'Free space of disk where param {self.option_name} is located is no greater than 1 GB, ' \
                         f'there is risk of disk exhaustion.'

            notice_accept_msg = 'Current login user has accepted the risk of disk exhaustion.'

            interaction_op = InteractionByEntrance(option_name=self.option_name,
                                                   option_value=self.option_value,
                                                   notice_msg=notice_msg,
                                                   notice_accept_msg=notice_accept_msg,
                                                   exception_type=LowDiskFreeSizeRiskError)
            interaction_op.interact_by_entrance(force_quit=force_quit, quiet=quiet)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) Huawei Technologies Co., Ltd. 2022-2023, All rights reserved.
"""Low Rank Adapter Cell"""
from mindspore import nn
import mindspore.common.dtype as mstype
from mindspore.common.initializer import initializer, Initializer
from mindspore.common.parameter import Parameter
from mindspore import Tensor
from mindspore.ops import operations as P
from mindspore.ops import functional as F
try:
    from mindspore.nn.transformer.layers import _args_type_validator_check, _valid_value_checks
    from mindspore._checkparam import Validator
except ImportError:
    import mindspore._checkparam as Validator
    from mindformers.modules.layers import _args_type_validator_check, _valid_value_checks
from mindpet.delta.delta_constants import VALID_TENSOR_DATATYPE
from mindpet.utils.version_control import get_activation, _activation

class LowRankLinear(nn.Cell):
    """
    定义微调算法LowRankLinear层, 其主要结构包含两个compactor。为LowRankAdapterLayer所调用。

    Args:
        in_channels (int): LowRankLinear输入Tensor的空间维度
        out_channels (int): LowRankLinear输出Tensor的空间维度
        rank: LowRankLinear的隐藏size
        weight_init (Union[str, Initializer]):
            权重参数的初始化方法。它的类型可以是str或Initializer。
            当使用str时，值引用自类initializer；更多细节请参考Initializer的值。
            默认值："normal"。
        bias_init (Union[Tensor, str, Initializer, numbers.Number]): 偏置参数的初始化方法。
            它的类型可以是Tensor，str，Initializer或numbers.Number。
            当使用str时，值引用自类initializer；更多细节请参考Initializer的值。
            当使用Tensor时，数据类型与输入Tensor相同。
            默认值："zeros"
        has_bias (int): LowRankLinear的权重矩阵是否有偏置
        param_init_type (dtype.Number): 表示dense中模块的参数初始化类型
        compute_dtype (dtype.Number): 表示dense中矩阵乘法的计算类型
    """

    @_args_type_validator_check(in_channels=Validator.check_positive_int,
                                out_channels=Validator.check_positive_int,
                                rank=Validator.check_positive_int,
                                has_bias=Validator.check_bool,
                                param_init_type=_valid_value_checks(VALID_TENSOR_DATATYPE,
                                                                    "LowRankLinear"),
                                compute_dtype=_valid_value_checks(VALID_TENSOR_DATATYPE,
                                                                  "LowRankLinear"))
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 rank: int = 1,
                 weight_init='normal',
                 bias_init='zeros',
                 has_bias: bool = True,
                 param_init_type: mstype = mstype.float32,
                 compute_dtype: mstype = mstype.float16):
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.rank = rank
        self.weight_init = weight_init
        self.mindpet_delta_low_rank_adapter_weight_left = \
            Parameter(initializer(self.weight_init, [in_channels, rank], param_init_type),
                      name="mindpet_delta_low_rank_adapter_weight_left")
        self.mindpet_delta_low_rank_adapter_weight_right = \
            Parameter(initializer(self.weight_init, [rank, out_channels], param_init_type),
                      name="mindpet_delta_low_rank_adapter_weight_right")
        self.has_bias = has_bias

        self.bias = None
        if self.has_bias:
            if isinstance(bias_init, Tensor):
                if bias_init.ndim != 1 or bias_init.shape[0] != out_channels:
                    raise ValueError(f"For '{self.cls_name}', bias init shape error. The ndim of 'bias_init' must "
                                     f"be equal to 1, and the first dim must be equal to 'out_channels'. But got "
                                     f"'bias_init': {bias_init}, 'out_channels': {out_channels}.")
            self.bias = Parameter(initializer(
                bias_init, [out_channels], param_init_type), name="bias")
            self.bias_add = P.BiasAdd()

        self.compt_dtype = compute_dtype
        self.matmul_weight = P.MatMul(transpose_b=False)
        self.matmul_input = P.MatMul(transpose_b=False)
        self.reshape = P.Reshape()
        self.shape_op = P.Shape()

    def construct(self, input_tensor):
        """Forward"""
        # get input_x info
        x_shape = self.shape_op(input_tensor)
        ori_dtype = F.dtype(input_tensor)

        # reshape input_tensor to compute
        input_tensor = self.reshape(input_tensor, (-1, x_shape[-1]))

        # compute weight
        weight = self.matmul_weight(self.cast(self.mindpet_delta_low_rank_adapter_weight_left, self.compt_dtype),
                                    self.cast(self.mindpet_delta_low_rank_adapter_weight_right, self.compt_dtype))

        input_tensor = self.cast(input_tensor, self.compt_dtype)
        input_tensor = self.matmul_input(input_tensor, weight)

        if self.has_bias:
            input_tensor = self.bias_add(
                input_tensor, self.cast(self.bias, self.compt_dtype))

        out_shape = x_shape[:-1] + (-1,)
        input_tensor = self.reshape(input_tensor, out_shape)

        output = self.cast(input_tensor, ori_dtype)
        return output

    def shard(self, strategy_matmul_weight, strategy_matmul_input, strategy_bias=None):
        self.matmul_weight.shard(strategy_matmul_weight)
        self.matmul_input.shard(strategy_matmul_input)
        if self.has_bias:
            self.bias_add.shard(strategy_bias)
        return self


class LowRankAdapterLayer(nn.Cell):
    """
    定义微调算法LowRankAdapter模块结构, 初始化LowRankAdapter层的参数, 包括矩阵参数、激活层等。

    Args:
        hidden_size (int): 隐藏层输出的特征向量维度，亦是LowRankAdapter模块输入的特征向量维度。
        reduction_factor (int): Low-Rank Adapter结构向下投影的维度值缩减倍数。
            如计算得bottleneck_dim = hidden_size//reduction_factor，
            bottleneck_dim即为向下投影的维度值。
            默认值为1。
        low_rank_size (int): LowRankAdapter模块中'内层'bottleneck的隐藏大小。
        low_rank_w_init (Union[str, Initializer, None]):
            Low-Rank Adapter结构向下或向上投影的权重参数的初始化方法。它可以是str，Initializer。
            当使用str时，值引用自类initializer；更多细节请参考Initializer的值。
            默认值："xavier_uniform"。
        non_linearity (str): Low-Rank Adapter结构中投影后所使用的激活函数。
            str的值引用自mindspore中 `get_activation` 方法所支持的激活函数类型。
            默认为 'gelu'
        param_init_type (dtype.Number): 表示Low-Rank Adapter结构的参数初始化类型。
            其值应为dtype.float32或dtype.float16。
            默认值：dtype.float32。
        compute_dtype (dtype.Number): 表示Low-Rank Adapter结构的矩阵乘法的计算类型。
            其值应为dtype.float32或dtype.float16。
            默认值：dtype.float16。

    Inputs:
        input_tensor (Tensor): 网络的所有输入组成的元组，shape为(*, in_channels)的Tensor，in_channels与入参中的hidden_size类型一致。
    Outputs:
        output (Tensor): LowRankAdapter dense的计算结果，shape为 (*, out_channels)，out_channels与入参中的hidden_size类型一致。
    """

    @_args_type_validator_check(hidden_size=Validator.check_positive_int,
                                reduction_factor=Validator.check_positive_int,
                                low_rank_size=Validator.check_positive_int,
                                non_linearity=_valid_value_checks(list(_activation.keys()),
                                                                  "LowRankAdapterLayer"),
                                param_init_type=_valid_value_checks(VALID_TENSOR_DATATYPE,
                                                                    "LowRankAdapterLayer"),
                                compute_dtype=_valid_value_checks(VALID_TENSOR_DATATYPE,
                                                                  "LowRankAdapterLayer"))
    def __init__(
            self,
            hidden_size: int,
            reduction_factor: int,
            low_rank_size: int = 1,
            low_rank_w_init="xavier_uniform",
            non_linearity: str = "gelu",
            param_init_type: mstype = mstype.float32,
            compute_dtype: mstype = mstype.float16):
        self._check_low_rank_w_init(low_rank_w_init)
        self._check_reduction_factor(hidden_size, reduction_factor)
        super().__init__()

        self.bottleneck_size = hidden_size // reduction_factor
        self.non_linearity = non_linearity

        self.mindpet_delta_low_rank_adapter_down_sampler = LowRankLinear(in_channels=hidden_size,
                                                                    out_channels=self.bottleneck_size,
                                                                    rank=low_rank_size,
                                                                    weight_init=low_rank_w_init,
                                                                    param_init_type=param_init_type,
                                                                    compute_dtype=compute_dtype)
        self.mindpet_delta_low_rank_adapter_non_linear = get_activation(
            non_linearity)
        self.mindpet_delta_low_rank_adapter_up_sampler = LowRankLinear(in_channels=self.bottleneck_size,
                                                                  out_channels=hidden_size,
                                                                  rank=low_rank_size,
                                                                  weight_init=low_rank_w_init,
                                                                  param_init_type=param_init_type,
                                                                  compute_dtype=compute_dtype)
        self.residual_add = P.Add()

    def construct(self, input_tensor):
        """Forward"""
        # get input_x info
        x_shape = P.Shape()(input_tensor)
        ori_dtype = F.dtype(input_tensor)

        # reshape input_tensor to compute
        input_tensor = P.Reshape()(input_tensor, (-1, x_shape[-1]))

        # calculate adapter_out
        adapter_down_sampler_output = self.mindpet_delta_low_rank_adapter_down_sampler(
            input_tensor)
        adapter_non_linear_output = self.mindpet_delta_low_rank_adapter_non_linear(
            adapter_down_sampler_output)
        adapter_output = self.mindpet_delta_low_rank_adapter_up_sampler(
            adapter_non_linear_output)

        # residual connection, add input and adapter_output
        input_tensor = self.residual_add(input_tensor, adapter_output)

        # recover the previous outshape and dtype
        out_shape = x_shape[:-1] + (-1,)
        input_tensor = P.Reshape()(input_tensor, out_shape)
        output = P.Cast()(input_tensor, ori_dtype)
        return output

    def shard(self, strategy_matmul_down_sampler_weight=None,
              strategy_matmul_down_sampler_input=None,
              strategy_bias_down_sampler=None,
              strategy_non_linearity=None,
              strategy_matmul_up_sampler_weight=None,
              strategy_matmul_up_sampler_input=None,
              strategy_bias_up_sampler=None,
              strategy_residual_add=None):
        """
        Set the shard for the LowRankAdapterLayer. All the strategy size should be equal to the inputs.

        Note:
            It is valid only in semi auto parallel or auto parallel mode.
            In other parallel modes, strategies set here will be ignored.

        Args:
            strategy_matmul_down_sampler_weight (tuple): The strategy for the
                LowRankAdapter-down_sampler-matmul_weight matmul.
            strategy_matmul_down_sampler_input (tuple): The strategy for the
                LowRankAdapter-down_sampler-matmul_input matmul.
            strategy_bias_down_sampler (tuple): The strategy for the
                LowRankAdapter-down_sampler-bias_add.
            strategy_non_linearity (tuple): The strategy for the LowRankAdapter non_linearity.
            strategy_matmul_up_sampler_weight (tuple): The strategy for the
                LowRankAdapter-up_sampler-matmul_weight matmul.
            strategy_matmul_up_sampler_input (tuple): The strategy for the
                LowRankAdapter-up_sampler-matmul_input matmul.
            strategy_bias_up_sampler (tuple): The strategy for the
                LowRankAdapter-up_sampler-bias_add.
            strategy_residual_add (tuple): The strategy for the residual_add.
        """
        try:
            self.mindpet_delta_low_rank_adapter_down_sampler.shard(
                strategy_matmul_down_sampler_weight, strategy_matmul_down_sampler_input, strategy_bias_down_sampler)
            self.mindpet_delta_low_rank_adapter_up_sampler.shard(
                strategy_matmul_up_sampler_weight, strategy_matmul_up_sampler_input, strategy_bias_up_sampler)
            # some operations has many primitives, need to manually set the shard
            if self.non_linearity.lower() == "leakyrelu":
                self.mindpet_delta_low_rank_adapter_non_linear.select_op.shard(
                    (strategy_non_linearity[0], strategy_non_linearity[0]))
            elif self.non_linearity.lower() == "logsigmoid":
                self.mindpet_delta_low_rank_adapter_non_linear.mul.shard(
                    (strategy_non_linearity[0], ()))
                self.mindpet_delta_low_rank_adapter_non_linear.exp.shard(
                    strategy_non_linearity)
                self.mindpet_delta_low_rank_adapter_non_linear.add.shard(
                    (strategy_non_linearity[0], ()))
                self.mindpet_delta_low_rank_adapter_non_linear.rec.shard(
                    strategy_non_linearity)
                self.mindpet_delta_low_rank_adapter_non_linear.log.shard(
                    strategy_non_linearity)
            elif self.non_linearity.lower() == "logsoftmax":
                raise ValueError("The 'LogSoftmax' function is not supported in semi auto parallel "
                                 "or auto parallel mode.")
            else:
                getattr(self.mindpet_delta_low_rank_adapter_non_linear,
                        self.non_linearity).shard(strategy_non_linearity)
            self.residual_add.shard(strategy_residual_add)

        except Exception as ex:
            # pylint: disable=W0719
            raise Exception(
                f"Exception occurred when set the shard for LowRankAdapterLayer, error message: {str(ex)}") from ex

        return self

    def _check_low_rank_w_init(self, low_rank_w_init):
        if not isinstance(low_rank_w_init, (str, Initializer)):
            raise TypeError(f"For '{self.cls_name}', the type of the 'low_rank_w_init' argument should be 'string' "
                            f"or 'initializer', but got {type(low_rank_w_init)}.")

    def _check_reduction_factor(self, hidden_size, reduction_factor):
        if reduction_factor > hidden_size:
            raise ValueError(f"For '{self.cls_name}', the value of the 'reduction_factor' argument should be smaller "
                             f"than 'hidden_size', but got reduction_factor: {reduction_factor}, "
                             f"hidden_size: {hidden_size}.")


class LowRankAdapterDense(nn.Dense):
    """
    定义微调算法LowRankAdapter dense层，继承nn.Dense。

    Args:
        in_channels (int): LowRankAdapterDense层输入Tensor的空间维度。
        out_channels (int): LowRankAdapterDense层输出Tensor的空间维度。
        weight_init (Union[Tensor, str, Initializer, numbers.Number]):
            LowRankAdapterDense中全连接层权重参数的初始化方法。
            它的类型可以是Tensor，str，Initializer或numbers.Number。
            当使用str时，值引用自类initializer；更多细节请参考Initializer的值。
            当使用Tensor时，数据类型与输入Tensor相同。
            默认值："normal"。
        bias_init (Union[Tensor, str, Initializer, numbers.Number]):
            LowRankAdapterDense中全连接层偏置参数的初始化方法。
            它的类型可以是Tensor，str，Initializer或numbers.Number。
            当使用str时，值引用自类initializer；更多细节请参考Initializer的值。
            当使用Tensor时，数据类型与输入Tensor相同。
            默认值："zeros"。
        has_bias (bool): 是否有偏置
        activation (Union[str, Cell, Primitive, None]): 激活函数，是与创建层的输入具有相同数据类型的权重矩阵
        reduction_factor (int): bottleneck的瓶颈倍数，bottleneck_dim = hidden_size//reduction_factor
        low_rank_size (int): LowRankAdapter模块中内层bottleneck的隐藏大小
        low_rank_w_init (Union[str, Initializer, None]):
            Low-Rank Adapter结构向下或向上投影的权重参数的初始化方法。它可以是str或Initializer。
            当使用str时，值引用自类initializer；更多细节请参考Initializer的值。
            默认值："xavier_uniform"。
        non_linearity (str): Low-Rank Adapter结构中投影后所使用的激活函数。
            str的值引用自mindspore中 `get_activation` 方法所支持的激活函数类型。
            默认为 'gelu'。
        param_init_type (dtype.Number): 表示dense中Low-Rank Adapter结构的参数初始化类型。
            其值应为dtype.float32或dtype.float16。
            默认值：dtype.float32。
        compute_dtype (dtype.Number): 表示dense中Low-Rank Adapter结构的矩阵乘法的计算类型。
            其值应为dtype.float32或dtype.float16。
            默认值：dtype.float16。

    Inputs:
        input_tensor (Tensor): 网络的所有输入组成的元组，shape为(*, in_channels)的Tensor，in_channels与入参中的in_channels类型一致。
    Outputs:
        output (Tensor): LowRankAdapter dense的计算结果，shape为 (*, out_channels)，out_channels与入参中的out_channels类型一致。
    """

    @_args_type_validator_check(in_channels=Validator.check_positive_int,
                                out_channels=Validator.check_positive_int,
                                has_bias=Validator.check_bool,
                                reduction_factor=Validator.check_positive_int,
                                low_rank_size=Validator.check_positive_int,
                                non_linearity=_valid_value_checks(list(_activation.keys()),
                                                                  "LowRankAdapterDense"),
                                param_init_type=_valid_value_checks(VALID_TENSOR_DATATYPE,
                                                                    "LowRankAdapterDense"),
                                compute_dtype=_valid_value_checks(VALID_TENSOR_DATATYPE,
                                                                  "LowRankAdapterDense"))
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 weight_init='normal',
                 bias_init='zeros',
                 has_bias=True,
                 activation=None,
                 reduction_factor: int = 1,
                 low_rank_size: int = 1,
                 low_rank_w_init="xavier_uniform",
                 non_linearity: str = "gelu",
                 param_init_type: mstype = mstype.float32,
                 compute_dtype: mstype = mstype.float16):
        super().__init__(in_channels=in_channels,
                         out_channels=out_channels,
                         weight_init=weight_init,
                         bias_init=bias_init,
                         has_bias=has_bias,
                         activation=activation)
        self.mindpet_delta_low_rank_adapter = LowRankAdapterLayer(hidden_size=out_channels,
                                                             reduction_factor=reduction_factor,
                                                             low_rank_size=low_rank_size,
                                                             low_rank_w_init=low_rank_w_init,
                                                             non_linearity=non_linearity,
                                                             param_init_type=param_init_type,
                                                             compute_dtype=compute_dtype, )
        self.compt_dtype = compute_dtype
        self.cast = P.Cast()
        self.act_name = activation

    def construct(self, input_tensor):
        """Forward"""
        # get input_x info
        x_shape = self.shape_op(input_tensor)
        ori_dtype = F.dtype(input_tensor)

        # reshape input_tensor to compute
        input_tensor = self.reshape(input_tensor, (-1, x_shape[-1]))

        # start to linear compute
        weight = self.cast(self.weight, self.compt_dtype)
        input_tensor = self.cast(input_tensor, self.compt_dtype)

        input_tensor = self.matmul(input_tensor, weight)
        if self.has_bias:
            input_tensor = self.bias_add(
                input_tensor, self.cast(self.bias, self.compt_dtype))
        if self.activation_flag:
            input_tensor = self.activation(input_tensor)

        # calculate low_rank_adapter_out
        input_tensor = self.mindpet_delta_low_rank_adapter(input_tensor)

        # recover the previous outshape and dtype
        out_shape = x_shape[:-1] + (-1,)
        input_tensor = self.reshape(input_tensor, out_shape)
        output = self.cast(input_tensor, ori_dtype)
        return output

    def shard(self, strategy_matmul_org=None,
              strategy_bias_org=None,
              strategy_activation_org=None,
              strategy_matmul_down_sampler_weight=None,
              strategy_matmul_down_sampler_input=None,
              strategy_bias_down_sampler=None,
              strategy_non_linearity=None,
              strategy_matmul_up_sampler_weight=None,
              strategy_matmul_up_sampler_input=None,
              strategy_bias_up_sampler=None,
              strategy_residual_add=None):
        """
        Set the shard for the LowRankAdapterDense. All the strategy size should be equal to the inputs.

        Note:
            It is valid only in semi auto parallel or auto parallel mode.
            In other parallel modes, strategies set here will be ignored.

        Args:
            strategy_matmul_org (tuple): The strategy for the origin dense matmul.
            strategy_bias_org (tuple): The strategy for the origin dense bias_add.
            strategy_activation_org (tuple): The strategy for the LowRankAdapterDense dropout.
            strategy_matmul_down_sampler_weight (tuple): The strategy for the
                LowRankAdapter-down_sampler-matmul_weight matmul.
            strategy_matmul_down_sampler_input (tuple): The strategy for the
                LowRankAdapter-down_sampler-matmul_input matmul.
            strategy_bias_down_sampler (tuple): The strategy for the
                LowRankAdapter-down_sampler-bias_add.
            strategy_non_linearity (tuple): The strategy for the LowRankAdapter non_linearity.
            strategy_matmul_up_sampler_weight (tuple): The strategy for the
                LowRankAdapter-up_sampler-matmul_weight matmul.
            strategy_matmul_up_sampler_input (tuple): The strategy for the
                LowRankAdapter-up_sampler-matmul_input matmul.
            strategy_bias_up_sampler (tuple): The strategy for the
                LowRankAdapter-up_sampler-bias_add.
            strategy_residual_add (tuple): The strategy for the residual_add.
        """
        try:
            # set origin dense strategy
            self.matmul.shard(strategy_matmul_org)
            if self.has_bias:
                self.bias_add.shard(strategy_bias_org)
            if self.activation_flag and isinstance(self.act_name, str):
                if self.act_name.lower() == "leakyrelu":
                    self.activation.select_op.shard(
                        (strategy_activation_org[0], strategy_activation_org[0]))
                elif self.act_name.lower() == "logsigmoid":
                    self.activation.mul.shard((strategy_activation_org[0], ()))
                    self.activation.exp.shard(strategy_activation_org)
                    self.activation.add.shard((strategy_activation_org[0], ()))
                    self.activation.rec.shard(strategy_activation_org)
                    self.activation.log.shard(strategy_activation_org)
                elif self.act_name.lower() == "logsoftmax":
                    raise ValueError("The 'LogSoftmax' function is not supported in semi auto parallel "
                                     "or auto parallel mode.")
                else:
                    getattr(self.activation, self.act_name).shard(
                        strategy_activation_org)

            # set low_rank_adapter strategy
            self.mindpet_delta_low_rank_adapter.shard(strategy_matmul_down_sampler_weight,
                                                 strategy_matmul_down_sampler_input,
                                                 strategy_bias_down_sampler,
                                                 strategy_non_linearity,
                                                 strategy_matmul_up_sampler_weight,
                                                 strategy_matmul_up_sampler_input,
                                                 strategy_bias_up_sampler,
                                                 strategy_residual_add)
        except Exception as ex:
            # pylint: disable=W0719
            raise Exception(
                f"Exception occurred when set the shard for LowRankAdapterDense, error message: {str(ex)}") from ex

        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""Entrance Monitor"""
class EntranceMonitor:
    """Entrance Monitor"""
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = object.__new__(cls)
        return cls._instance

    @staticmethod
    def init():
        global MONITOR_DICT
        MONITOR_DICT = {}

    @staticmethod
    def set_value(key, value):
        MONITOR_DICT[key] = value

    @staticmethod
    def get_value(key):
        return MONITOR_DICT.get(key)


MONITOR_DICT = {}
entrance_monitor = EntranceMonitor()
entrance_monitor.init()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved."""

from mindpet.utils.constants import GB_SIZE
from mindpet.security.param_check.base_check import BaseCheckParam, BaseCheck
from mindpet.security.param_check.option_check_utils import PathContentBlacklistCharactersCheck, AbsolutePathCheck, \
    PathExistCheck, LinkPathCheck, PathContentLengthCheck, PathContentCharacterCheck, PathGranularityCheck, \
    DiskFreeSpaceCheck, get_real_path


class FinetuneOptionsCheckParam(BaseCheckParam):
    """class FinetuneOptionsCheckParam class"""
    def __init__(self,
                 path_length_check_param,
                 path_content_check_param,
                 mode,
                 path_including_file=False,
                 force_quit=False,
                 quiet=False):
        """
        微调入参校验参数
        :param path_length_check_param: 路径长度校验参数
        :param path_content_check_param: 路径内容校验参数
        :param mode: 路径权限范围约束
        :param path_including_file: 路径是否为文件路径
        :param force_quit: SDK场景下, 路径权限提升校验异常时是否强制退出
        :param quiet: CLI场景下，路径权限提升校验是否忽略交互式过程
        """
        super().__init__(mode, path_including_file, force_quit, quiet)
        self.path_length_check_param = path_length_check_param
        self.path_content_check_param = path_content_check_param


class FinetuneOptionsCheck(BaseCheck):
    """FinetuneOptionsCheck class"""
    def __init__(self, option_name, option_value, disk_space_check=False):
        super().__init__(option_name, option_value)
        self.disk_space_check = disk_space_check

    def check(self, check_param):
        """
        微调入参校验
        :param check_param: 微调入参校验参数
        """
        # 针对原始路径进行的校验
        self._origin_path_check_item()

        # 获取真实路径
        self.option_value = get_real_path(self.option_value)

        # 针对真实路径进行的校验
        self._real_path_check_item(check_param)

    def _origin_path_check_item(self):
        """_origin_path_check_item"""
        # 路径内容黑名单校验
        PathContentBlacklistCharactersCheck(option_name=self.option_name, option_value=self.option_value)

        # 绝对路径校验
        AbsolutePathCheck(option_name=self.option_name, option_value=self.option_value)

        # 路径真实性校验
        PathExistCheck(option_name=self.option_name, option_value=self.option_value)

        # 路径软链接校验
        LinkPathCheck(option_name=self.option_name, option_value=self.option_value)

    def _real_path_check_item(self, check_param):
        """_real_path_check_item"""
        # 路径长度校验
        PathContentLengthCheck(option_name=self.option_name, option_value=self.option_value,
                               path_min_limit=check_param.path_length_check_param.path_min_limit,
                               path_max_limit=check_param.path_length_check_param.path_max_limit,
                               file_min_limit=check_param.path_length_check_param.file_min_limit,
                               file_max_limit=check_param.path_length_check_param.file_max_limit)

        # 路径内容白名单校验
        PathContentCharacterCheck(option_name=self.option_name, option_value=self.option_value,
                                  base_whitelist_mode=check_param.path_content_check_param.base_whitelist_mode,
                                  extra_whitelist=check_param.path_content_check_param.extra_whitelist)

        # 路径粒度校验
        PathGranularityCheck(option_name=self.option_name, option_value=self.option_value,
                             path_including_file=check_param.path_including_file)

        self.base_check_item(check_param)

        # 路径所在磁盘空间耗尽风险校验
        if self.disk_space_check:
            DiskFreeSpaceCheck(option_name=self.option_name,
                               option_value=self.option_value,
                               free_space_limit=GB_SIZE,
                               force_quit=check_param.force_quit,
                               quiet=check_param.quiet)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""timeout option"""

import re
import click
from click.exceptions import BadParameter
from mindpet.utils.constants import TIMEOUT_REGEX, HOURS_PER_DAY, SPACE_CHARACTER, EMPTY_STRING, SECONDS_PER_HOUR


class TimeoutOption(click.core.Option):
    """TimeoutOption class"""
    def __init__(self):
        super().__init__(
            param_decls=('-t', '--timeout'),
            type=str,
            default=None,
            help='if the task exceeds the time limit, it will be forcibly terminated.',
            callback=self.timeout_callback
        )

    @staticmethod
    def get_timeout_hours(match_result):
        """get_timeout_hours"""
        res = 0

        extract_days_str = match_result.group(1)
        if extract_days_str:
            res += (int(extract_days_str.strip('d')) * HOURS_PER_DAY)

        extract_hours_str = match_result.group(2)
        if extract_hours_str:
            extract_hours = int(extract_hours_str.strip('h'))
            if extract_hours >= HOURS_PER_DAY:
                raise BadParameter('Invalid param [timeout].')

            res += extract_hours

        if res == 0:
            raise BadParameter('Invalid param [timeout].')

        res *= SECONDS_PER_HOUR

        return res

    # pylint: disable=W0613
    def timeout_callback(self, ctx, param, value):
        """
        timeout参数click回调方法

        :param ctx: 上下文信息
        :param param: 参数属性
        :param value: 输入值
        :return: 回调处理后的参数值
        """
        if value is None:
            return value

        value = value.strip().replace(SPACE_CHARACTER, EMPTY_STRING).lower()
        match_result = re.match(TIMEOUT_REGEX, value)

        if not match_result:
            raise BadParameter('Invalid param [timeout].')

        # 将输入的timeout参数转化为秒数
        res = self.get_timeout_hours(match_result)

        return res



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.

"""
定义冻结功能相关接口
"""

from typing import Optional, List
from fnmatch import fnmatch
from mindspore import nn

from mindpet.log.log import logger
from mindpet.utils.constants import DELTA_LIST
from mindpet.utils.exceptions import ReadYamlFileError, ModelConfigFreezeInfoError, LinkPathError, UnexpectedError
from mindpet.utils.io_utils import read_yaml_file


def freeze_modules(model: nn.Cell,
                   include: Optional[List[str]] = None,
                   exclude: Optional[List[str]] = None) -> None:
    """
    根据指定模块列表冻结网络。

    :param model: 需要冻结的模型实例， 必填。
    :param include: 需要冻结的模块名列表， 选填。
                    模糊匹配列表中所有模块名，挨个将匹配到的模块的requires_grad设置为False。
                    列表项支持配置符号*，代表任意字符串，格式如 ['*', '*dense*'，'*.dense.*', '*.dense.*.bias']
    :param exclude: 不冻结的模块名列表， 选填。
                    模糊匹配列表中所有模块名，挨个将匹配到的模块的requires_grad设置为True。
                    列表项支持配置符号*，代表任意字符串，格式如 ['*', '*dense*'，'*.dense.*', '*.dense.*.bias']
                    当include和exclude列表项冲突时，对该项匹配到的模块不做任何处理。
    """
    logger.info("Start to freeze model, include list: %s, exclude list: %s", include, exclude)

    if not isinstance(model, nn.Cell):
        raise TypeError("A Cell is required for argument 'model'.")

    if include is None and exclude is None:
        raise ValueError("Argument 'include' (pos 2) and 'exclude' (pos 3) can't be None together.")

    if include is not None and (not isinstance(include, List) or not include):
        raise TypeError("A non-empty list is required for argument 'include' (pos 2).")

    if exclude is not None and (not isinstance(exclude, List) or not exclude):
        raise TypeError("A non-empty list is required for argument 'exclude' (pos 3).")

    _freeze_from_list(model, include, exclude)
    logger.info("End to freeze model.")


def _freeze_from_list(model, include, exclude):
    """
    根据include/exclude列表冻结网络。
    """
    for name, param in model.parameters_and_names():
        if _match_str_and_list(name, include) and not _match_str_and_list(name, exclude):
            param.requires_grad = False
        elif not _match_str_and_list(name, include) and _match_str_and_list(name, exclude):
            param.requires_grad = True


def freeze_delta(model: nn.Cell,
                 mode: str,
                 include: Optional[List[str]] = None,
                 exclude: Optional[List[str]] = None) -> None:
    """
    根据微调算法类型以及指定模块列表冻结网络。
    目前已实现lora和prefixtuning两种微调算法的冻结模式。

    :param model: 需要冻结的模型实例，必填。
    :param mode: 微调算法类型，必填。可选填'lora'或'prefixtuning'。
    :param include: 需要冻结的模块名列表， 选填。
                    模糊匹配列表中所有模块名，挨个将匹配到的模块的requires_grad设置为False。
                    列表项支持配置符号*，代表任意字符串，格式如 ['*', '*dense*'，'*.dense.*', '*.dense.*.bias']
    :param exclude: 不冻结的模块名列表， 选填。
                    模糊匹配列表中所有模块名，挨个将匹配到的模块的requires_grad设置为True。
                    列表项支持配置符号*，代表任意字符串，格式如 ['*', '*dense*'，'*.dense.*', '*.dense.*.bias']
                    当include和exclude列表项冲突时，对该项匹配到的模块不做任何处理。
    """
    logger.info("Start to freeze model for delta, mode: %s, include list: %s, exclude list: %s",
                mode, include, exclude)
    if not isinstance(model, nn.Cell):
        raise TypeError("A Cell is required for argument 'model'.")

    if isinstance(mode, str) and mode.lower() in DELTA_LIST:
        _freeze_for_mode(model, mode)
    else:
        raise ValueError(f"An argument of string type from {DELTA_LIST} is required.")

    if include or exclude:
        try:
            freeze_modules(model, include, exclude)
        except Exception as ex:
            raise UnexpectedError(f"Exception occurred when freeze model for delta, error message: {str(ex)}") from ex

    logger.info("End to freeze model for delta.")


def freeze_from_config(model: nn.Cell,
                       config_path: str) -> None:
    """
    根据配置文件中freeze关键词下的include和exclude列表冻结网络指定模块。

    :param model: 需要冻结的模型实例，必填。
    :param config_path: 配置文件路径，必填。

    配置文件内容实例：
    freeze:
      include: ['*', '*dense*'，'*.dense.*', '*.dense.*.bias']
      exclude: ['*', '*dense*'，'*.dense.*', '*.dense.*.bias']

    include: 需要冻结的模块名列表， 选填。
            模糊匹配列表中所有模块名，挨个将匹配到的模块的requires_grad设置为False。
            列表项支持配置符号*，代表任意字符串，格式如 ['*', '*dense*'，'*.dense.*', '*.dense.*.bias']
    exclude: 不冻结的模块名列表， 选填。
            模糊匹配列表中所有模块名，挨个将匹配到的模块的requires_grad设置为True。
            列表项支持配置符号*，代表任意字符串，格式如 ['*', '*dense*'，'*.dense.*', '*.dense.*.bias']
            当include和exclude列表项冲突时，对该项匹配到的模块不做任何处理。
    """
    logger.info("Start to freeze model from config file.")
    if not isinstance(model, nn.Cell):
        raise TypeError("A Cell is required for argument 'model'.")

    if not isinstance(config_path, str):
        raise TypeError("A string is required for argument 'config_path'.")

    if not config_path:
        raise ValueError("Argument 'config_path' is empty.")

    try:
        content_dict = read_yaml_file(config_path)
    except (FileNotFoundError, LinkPathError) as ex:
        raise ReadYamlFileError(f'Error occurred when freeze model from config, error message: {str(ex)}') from ex
    except Exception as ex:
        raise UnexpectedError(f'Unexpected error occurred when freeze model from config, '
                              f'error message: {str(ex)}') from ex

    include, exclude = _get_freeze_modules_list(content_dict)
    freeze_modules(model, include, exclude)
    logger.info("End to freeze model from config file.")


def _freeze_for_mode(model: nn.Cell, mode: str) -> None:
    """
    根据微调算法冻结网络。

    :param model: 需要冻结的模型实例，必填。
    """
    delta_name = '*mindpet_delta_' + mode + '*'
    if mode == 'bitfit':
        delta_name = '*bias'
    freeze_modules(model, include=['*'], exclude=[delta_name])


def _get_freeze_modules_list(content_dict) -> tuple:
    """
    从配置文件中获取include和exclude列表。

    :param content_dict: 配置文件内容字典
    """
    if not content_dict or 'freeze' not in content_dict.keys():
        raise ModelConfigFreezeInfoError("Missing required key 'freeze' in config file.")

    freeze_dict = content_dict.get('freeze')
    if not isinstance(freeze_dict, dict):
        raise ModelConfigFreezeInfoError("A dict is required for 'freeze' value in config file.")

    if freeze_dict is None:
        raise ModelConfigFreezeInfoError("'freeze' value in config file is empty.")

    # 校验freeze下是否有include和exclude字典，以及是否有其它非法字典
    for key in freeze_dict.keys():
        if key not in ['include', 'exclude']:
            raise ModelConfigFreezeInfoError("Invalid key in 'freeze' value, just config 'include'/'exclude' key.")

    include_list = []
    exclude_list = []
    if 'include' in freeze_dict.keys():
        include_list = freeze_dict.get('include')

    if 'exclude' in freeze_dict.keys():
        exclude_list = freeze_dict.get('exclude')

    return include_list, exclude_list


def _match_str_and_list(m_str, m_list: Optional[List[str]] = None) -> bool:
    """
    校验字符串是否与列表中某一项匹配。
    :param m_str: 完整字符串
    :param m_list: 有关键词的列表
    """
    if m_list is None:
        return False

    for key in m_list:
        if not isinstance(key, str):
            raise TypeError(f"List item '{key}' is not a string.")

        if fnmatch(m_str, key):
            return True

    return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved."""

import os
import json
import subprocess
from mindpet.utils.io_utils import read_yaml_file, read_json_file
from mindpet.security.param_check.option_check_utils import PathContentCheckParam
from mindpet.task.evaluate_infer.result_file_check import ResultFileCheckParam, ResultFileCheck
from mindpet.log.log import logger, operation_logger_without_std, record_operation_and_service_info_log, \
    record_operation_and_service_error_log, record_operation_and_service_warning_log
from mindpet.utils.constants import MODEL_CONFIG_ROOT_KEYS, EVAL_RESULT_FILE_NAME, INFER_RESULT_FILE_NAME, SPACE_CHARACTER, \
    PATH_MODE_LIMIT, EVAL_INFER_TASK_NAMES, EVALUATE_TASK_NAME, EMPTY_STRING
from mindpet.utils.task_utils import create_output_path_subdir_with_uuid, model_config_params_check_item, \
    monitor_process_rsp_code, model_config_keys_check_item, path_in_out_loop_check
from mindpet.utils.exceptions import LinkPathError, ReadYamlFileError, UnexpectedError, MonitorProcessRspError, \
    CreateProcessError, TaskError, FileOversizeError, PathRightEscalationError


class EvaluateInferTask:
    """EvaluateInferTask class"""
    def __init__(self, task_type, *args, **kwargs):
        """
        评估/推理任务构造方法
        task_type: 任务类型(evaluate/infer)
        *args: 元组参数
        **kwargs: 字典参数
        """
        if task_type not in EVAL_INFER_TASK_NAMES:
            raise ValueError(f'Invalid task_type, only support [{EVAL_INFER_TASK_NAMES}].')

        self.task_type = task_type
        self.result_name = EVAL_RESULT_FILE_NAME if self.task_type == EVALUATE_TASK_NAME else INFER_RESULT_FILE_NAME
        self.args = args
        self.kwargs = kwargs
        self.command = []
        self.command_params = []

        try:
            self._process_param_and_command()
        except KeyboardInterrupt as ex:
            record_operation_and_service_error_log(
                f'{str(self.task_type).capitalize()} task is terminated by current user, task has stopped and exited.')
            raise ex
        except Exception as ex:
            record_operation_and_service_error_log(f'{str(self.task_type).capitalize()} failed.')
            raise ex

    # pylint: disable=R1732
    def start(self):
        """
        启动命令
        :return: 评估/推理JSON形式结果
        """
        # 启动评估/推理任务
        record_operation_and_service_info_log(
            f'{str(self.task_type).capitalize()} task is running.')

        try:
            process = subprocess.Popen(self.command, env=os.environ, shell=False,
                                       stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        except KeyboardInterrupt as ex:
            record_operation_and_service_error_log(
                f'{str(self.task_type).capitalize()} task is terminated by current user, task has stopped and exited.')
            raise ex

        except Exception as ex:
            if ex is None or not str(ex):
                raise CreateProcessError(f'Exception occurred when creating {self.task_type} task process, '
                                         f'no error message available.') from ex
            raise CreateProcessError(f'Exception occurred when creating {self.task_type} task process, '
                                     f'error message: {str(ex)}.') from ex

        # 启动超时监听程序
        try:
            rsp_code = monitor_process_rsp_code(task_name=self.task_type, task_process=process,
                                                timeout=self.kwargs.get('timeout'))
        except KeyboardInterrupt as ex:
            record_operation_and_service_error_log(
                f'{str(self.task_type).capitalize()} task is terminated by current user, task has stopped and exited.')
            raise ex
        except Exception as ex:
            if ex is None or not str(ex):
                raise MonitorProcessRspError(f'Exception occurred when monitoring {self.task_type} task process, '
                                             f'no error message available.') from ex
            raise MonitorProcessRspError(f'Exception occurred when monitoring {self.task_type} task process, '
                                         f'error message: {str(ex)}.') from ex

        if rsp_code != 0:
            message = f'{str(self.task_type).capitalize()} failed.'
            operation_logger_without_std.error(message)
            raise TaskError(message)

        # 获取评估结果
        result = self._get_task_result()

        if result.get('status') == 0:
            record_operation_and_service_info_log(f'{str(self.task_type).capitalize()} successfully.')
        else:
            record_operation_and_service_warning_log(
                f'Completed {self.task_type} task, but failed to get {self.result_name} file.')
            logger.warning(result.get("error_message"))

        return result

    # pylint: disable=W0718
    def _get_task_result(self):
        """
        读取并返回用户落盘的评估/推理结果
        :return: json格式的评估/推理结果
        """

        # 读取用户落盘结果，默认状态为失败
        result = {"status": -1,
                  "task_result": EMPTY_STRING}

        # 校验用户代码生成的评估/推理结果文件合法性
        full_output_path = self.command_params[self.command_params.index('--output_path') + 1]
        result_path = os.path.join(full_output_path, self.result_name)
        error_message = self._get_check_result_file_error_msg(result_path)

        # 如果未通过结果文件生成校验，直接返回评估/推理结果
        if error_message:
            result["error_message"] = error_message
            return result

        # 读取并解析评估/推理结果, 读取成功以json串形式并返回，无error_message；读取失败返回空字符串，有相应error_message。
        try:
            task_result = read_json_file(result_path)
        except (json.JSONDecodeError, TypeError):
            # 文件为空但文件大小不为0、格式错误
            result["error_message"] = f'File {self.result_name} should follow JSON format.'
        # pylint: disable=W0703
        except Exception as ex:
            result["error_message"] = f'An error occurred during reading {self.result_name}: {str(ex)}'
        else:
            if task_result:
                result["status"] = 0
                result["task_result"] = task_result
            else:
                result["error_message"] = f'File {self.result_name} is empty.'

        return result

    # pylint: disable=W0718
    def _get_check_result_file_error_msg(self, result_path):
        """
        获取结果文件校验结果与对应error_message
        :result_path: 用户落盘结果文件路径
        :return:
           check_flag, 校验是否合格
           error_message：对应错误信息
        """
        error_message = EMPTY_STRING
        try:
            self._check_task_result(result_path)
        except FileNotFoundError:
            error_message = f'Cannot find {self.result_name} in [output_path].'
        except PathRightEscalationError:
            error_message = f'Permission denied when reading {self.result_name}.'
        except FileOversizeError:
            error_message = f'File {self.result_name} is too large.'
        except LinkPathError:
            error_message = f'Detect link path, reject reading file: {self.result_name}.'
        except ValueError:
            error_message = f'Invalid file: {self.result_name}.'
        # pylint: disable=W0703
        except Exception as ex:
            error_message = f'An error occurred during reading {self.result_name}: {str(ex)}'

        return error_message

    def _process_param_and_command(self):
        """
        评估/推理任务启动前的准备工作
        """
        # 命令行参数准备
        self._prepare_command_params()

        # 命令准备
        self._prepare_command()

    def _check_task_result(self, eval_result_path):
        """
        校验评估/推理结果文件合法性
        """
        base_whitelist_mode = 'ALL'
        extra_whitelist = ['.', '/', '-', '_', SPACE_CHARACTER]

        result_file_name = EVAL_RESULT_FILE_NAME if self.task_type == EVALUATE_TASK_NAME else INFER_RESULT_FILE_NAME
        result_file_path = ResultFileCheck(option_name=result_file_name, option_value=eval_result_path)

        path_content_check_param = PathContentCheckParam(base_whitelist_mode=base_whitelist_mode,
                                                         extra_whitelist=extra_whitelist)

        check_param = ResultFileCheckParam(path_content_check_param=path_content_check_param,
                                           mode=PATH_MODE_LIMIT,
                                           path_including_file=True,
                                           force_quit=False,
                                           quiet=self.kwargs.get('quiet'))

        result_file_path.check(check_param)

    def _prepare_command_params(self):
        """
        准备命令行参数
        """
        # 输入、输出回环路径判断
        path_in_out_loop_check(self.kwargs)

        self._wrap_model_config_params()
        self._wrap_required_params()

    def _wrap_required_params(self):
        """
        组装必填参数
        """
        output_path = create_output_path_subdir_with_uuid(self.kwargs.get('output_path'))

        self.command_params.extend(['--output_path', output_path])
        self.command_params.extend(['--data_path', self.kwargs.get('data_path')])
        self.command_params.extend(['--ckpt_path', self.kwargs.get('ckpt_path')])

    def _wrap_model_config_params(self):
        """
        针对model_config配置文件内容，组装相关参数
        """
        model_config_option = self.kwargs.get('model_config_path')

        if model_config_option is not None:
            try:
                content = read_yaml_file(model_config_option)
            except (FileNotFoundError, LinkPathError) as ex:
                raise ReadYamlFileError(
                    f'Error occurred when getting {self.task_type} params info from param [model_config_path], '
                    f'error message: {str(ex)}') from ex
            except Exception as ex:
                raise UnexpectedError(
                    f'Unexpected error occurred when getting {self.task_type} params info from '
                    f'param [model_config_path], error message: {str(ex)}') from ex

            model_config_keys_check_item(content, MODEL_CONFIG_ROOT_KEYS[0])
            model_config_params_check_item(task_object=self, content=content)

    def _prepare_command(self):
        """
        组装命令
        """
        self.command.append('python')
        self.command.append(self.kwargs.get('boot_file_path'))
        self.command.extend(self.command_params)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""Utils Constants"""
import os
import stat


# 空字符串
EMPTY_STRING = ''

# 入参本地路径黑名单字符过滤
BLACKLIST_CHARACTERS = (
    '\'', '"',
    '..', '../', '%2e%2e', '%2e./', '.%2e/', '..%2f', '%2e%2e/', '%2e.%2f', '.%2e%2f', '%2e%2e%2f'
)

# 入参选项字符粒度参数校验
SPACE_CHARACTER = ' '
UPPER_CASE_LETTERS = [chr(i) for i in range(65, 91)]
LOWER_CASE_LETTERS = [chr(i) for i in range(97, 123)]
NUMBERS = [chr(i) for i in range(48, 58)]
BASE_WHITELIST_CHARACTERS = {
    'UPPER': UPPER_CASE_LETTERS,
    'LOWER': LOWER_CASE_LETTERS,
    'LETTERS': UPPER_CASE_LETTERS + LOWER_CASE_LETTERS,
    'NUMBERS': NUMBERS,
    'ALL': UPPER_CASE_LETTERS + LOWER_CASE_LETTERS + NUMBERS
}

# 入参长度默认限制
DEFAULT_PATH_LEN_MIN_LIMIT = 1
DEFAULT_PATH_LEN_MAX_LIMIT = 4096
DEFAULT_FILE_LEN_MIN_LIMIT = 1
DEFAULT_FILE_LEN_MAX_LIMIT = 255

# 文件大小限制
MB_SIZE = 1024 * 1024
GB_SIZE = 1024 * 1024 * 1024
MAX_FILE_MB_SIZE = 10

# 文件I/O默认权限配置
DEFAULT_FLAGS = os.O_RDWR | os.O_CREAT  # 允许读写, 文件不存在时新建
DEFAULT_MODES = stat.S_IWUSR | stat.S_IRUSR  # 所有者读写

# finetune model_config配置文件相关配置
MODEL_CONFIG_ROOT_KEYS = ('params', 'freeze')
MODEL_CONFIG_PARAMS_CHARACTER_BLACKLIST = ('|', '&', ',', ';', '`', '"', '\'', '$', '(', ')', '{', '}', '>', '<', '?',
                                           '\\', '!', '\n', '*', '#')
MODEL_CONFIG_LEN_LIMIT = pow(2, 14)

# 微调基础包入口标识
ENTRANCE_TYPE = 'ENTRANCE_TYPE'

# 路径权限最小范围
MIN_MODE = 0
MAX_MODE = 7
PATH_MODE_LIMIT = '750'
DEFAULT_MAX_LOG_FILE_SIZE = 10
DEFAULT_MAX_LOG_FILE_NUM = 10

# output_path内部随机命名文件夹权限控制
OUTPUT_PATH_RANDOM_DIR_MODE = 0o750

# 评估结果文件名
EVAL_RESULT_FILE_NAME = 'eval_result.json'

# 推理结果文件名
INFER_RESULT_FILE_NAME = 'infer_result.json'

# 超时机制相关超参数
TIME_MONITOR_SLEEP_TIME = 1
TIMEOUT_REGEX = '^([0-9]+d)?([0-9]+h)?$'
SECONDS_PER_HOUR = 3600
HOURS_PER_DAY = 24

# 命令行模式异常退出状态码
ABNORMAL_EXIT_CODE = 1

# MindPet定义入参
MINDPET_DEFINED_PARAM_NAME = (
    'dp', 'data_path', 'op', 'output_path', 'bp', 'boot_file_path', 'mc', 'model_config_path', 'pm',
    'pretrained_model_path', 'cp', 'ckpt_path', 'q', 'quiet', 't', 'timeout', 'advanced_config')

# finetune接口入参顺序约束
ARG_NAMES = {
    'finetune': (
        'data_path', 'output_path', 'boot_file_path', 'pretrained_model_path', 'model_config_path', 'timeout'),
    'evaluate': ('data_path', 'ckpt_path', 'output_path', 'boot_file_path', 'model_config_path', 'timeout'),
    'infer': ('data_path', 'ckpt_path', 'output_path', 'boot_file_path', 'model_config_path', 'timeout')
}

# 自定义超参数不合法前缀
INVALID_CUSTOM_PARAM_KEY_PREFIX = '-'
INVALID_CUSTOM_PARAM_VAL_PREFIX = '--'

# 评估推理任务相关常量
FINETUNE_TASK_NAME = 'finetune'
EVALUATE_TASK_NAME = 'evaluate'
INFER_TASK_NAME = 'infer'
EVAL_INFER_TASK_NAMES = [EVALUATE_TASK_NAME, INFER_TASK_NAME]

# 微调工具包SDK接口清单
MINDPET_SDK_INTERFACE_NAMES = [FINETUNE_TASK_NAME, EVALUATE_TASK_NAME, INFER_TASK_NAME]

# 微调算法清单
DELTA_LIST = ['lora', 'prefixtuning', 'adapter', 'low_rank_adapter', 'bitfit', "ptuning2"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""model config path option"""

import click
from click.exceptions import BadParameter
from mindpet.utils.constants import PATH_MODE_LIMIT
from mindpet.task.finetune.finetune_options_check import FinetuneOptionsCheckParam, FinetuneOptionsCheck
from mindpet.task.options.path_check_param import default_file_length_check_param, default_file_content_check_param


class ModelConfigPathOption(click.core.Option):
    """ModelConfigPathOption class"""
    def __init__(self):
        super().__init__(
            param_decls=('-mc', '--model_config_path'),
            type=str,
            default=None,
            show_default=True,
            help='model config file local path',
            callback=self.model_config_path_callback
        )

    @staticmethod
    def _is_yaml_file(path):
        """
        路径是否是yaml文件
        :param path: 指定路径
        :return: 路径是否是yaml文件
        """
        return path.endswith('.yaml') or path.endswith('.yml')

    def model_config_path_callback(self, ctx, param, value):
        """
        model_config_path参数click回调方法

        :param ctx: 上下文信息
        :param param: 参数属性
        :param value: 输入值
        :return: 回调处理后的参数值
        """
        if value is None:
            return value

        check_param = FinetuneOptionsCheckParam(path_length_check_param=default_file_length_check_param,
                                                path_content_check_param=default_file_content_check_param,
                                                mode=PATH_MODE_LIMIT,
                                                path_including_file=True,
                                                quiet=ctx.params.get('quiet', False))
        model_config_path = FinetuneOptionsCheck(option_name=param.name, option_value=value)
        model_config_path.check(check_param)

        if not self._is_yaml_file(model_config_path.option_value):
            raise BadParameter('Param [model_config_path] only support .yaml or .yml file path.')

        return model_config_path.option_value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2010-2022. All rights reserved.
"""Prefix Layer Cell"""
import mindspore as ms
from mindspore import nn
try:
    from mindspore._checkparam import Validator, Rel
    INC_LEFT = Rel.INC_LEFT
except ImportError:
    import mindspore._checkparam as Validator
    INC_LEFT = Validator.INC_LEFT
from mindpet.utils.version_control import get_dropout


def check_multiple(param_dividend, value_dividend, param_divisor, value_divisor):
    if value_dividend % value_divisor != 0:
        raise ValueError(f"param {param_dividend} must be a multiple of {param_divisor}")
    return value_dividend / value_divisor


class PrefixLayer(nn.Cell):
    """A layer of prefix tuning module."""
    def __init__(self,
                 prefix_token_num,
                 batch_size,
                 num_heads,
                 hidden_dim,
                 embed_dim,
                 mid_dim=512,
                 dropout_rate=0.1):
        """

        Args:
            prefix_token_num: prefix的长度
            batch_size: 单次传递给程序用以训练的数据个数，与原模型参数一致
            num_heads: 多头注意力头数，与原模型参数一致
            hidden_dim: 模型隐藏维度， 与原模型参数一致
            embed_dim: embedding的维度， 与原模型参数一致
            mid_dim: prefix中间维度
            dropout_rate: 丢弃率
        """
        super().__init__()
        self.prefix_token_num = Validator.check_positive_int(prefix_token_num, int, "prefix_token_num")
        self.batch_size = Validator.check_positive_int(batch_size, int, "batch_size")
        self.num_heads = Validator.check_positive_int(num_heads, int, "num_heads")
        self.hidden_dim = Validator.check_positive_int(hidden_dim, int, "hidden_dim")
        self.embed_dim = Validator.check_positive_int(embed_dim, int, "embed_dim")
        self.mid_dim = Validator.check_positive_int(mid_dim, int, "mid_dim")
        self.dropout_rate = Validator.check_float_range(dropout_rate, 0.0, 1.0, INC_LEFT)
        try:
            check_multiple("prefix_token_num", prefix_token_num, "batch_size", batch_size)
        except ValueError as ex:
            raise ValueError(f"Invalid param [prefix_token_num] when initializing"
                             f"PrefixLayer, error message:{str(ex)}") from ex
        self.dropout = get_dropout(dropout_rate)
        self.past_value_reparam = None
        self.past_key_reparam = None
        self.__define_network()
        self.__allocate_parameter()

    def __define_network(self) -> None:
        """the network structure of prefix"""
        self.input_tokens = ms.Parameter(ms.numpy.arange(0, self.prefix_token_num, 1),
                                                name="mindpet_delta_prefixtuning_input_tokens", requires_grad=False)
        self.mindpet_delta_prefixtuning_wte = nn.Embedding(self.prefix_token_num, self.embed_dim)
        self.mindpet_delta_prefixtuning_control_trans = nn.SequentialCell(
            nn.Dense(self.embed_dim, self.mid_dim),
            nn.Tanh(),
            nn.Dense(self.mid_dim, self.hidden_dim)
        )

    def __allocate_parameter(self):
        """the value of prefix matrix"""
        input_tokens = self.input_tokens
        temp_control = self.mindpet_delta_prefixtuning_wte(input_tokens)
        past_key_values = self.mindpet_delta_prefixtuning_control_trans(temp_control)
        seq_len, _ = past_key_values.shape
        past_key_values = past_key_values.view(seq_len, -1, self.hidden_dim)
        past_key_values = self.dropout(past_key_values)
        past_key_values = ms.ops.transpose(past_key_values, (1, 0, 2)).split(2)
        self.past_key_reparam = past_key_values[0][0]
        self.past_value_reparam = past_key_values[0][0]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""MindPet Tuning Main."""

import sys
import click
from mindpet.utils.task_utils import handle_exception_log
from mindpet.utils.entrance_monitor import entrance_monitor
from mindpet.task.finetune.finetune_task import FinetuneTask
from mindpet.task.evaluate_infer.evaluate_infer_task import EvaluateInferTask
from mindpet.log.log import set_logger_property, record_operation_and_service_info_log
from mindpet.task.option_decorators import finetune_options, evaluate_options, infer_options
from mindpet.utils.constants import ENTRANCE_TYPE, EVALUATE_TASK_NAME, INFER_TASK_NAME, ABNORMAL_EXIT_CODE


def cli_wrapper():
    """
    CLI侧入口
    """
    entrance_monitor.set_value(ENTRANCE_TYPE, 'CLI')
    # 设置日志内容发起端标识
    set_logger_property("CLI")

    try:
        cli(standalone_mode=False)
    # pylint: disable=W0719
    # pylint: disable=W0703
    except Exception as ex:
        handle_exception_log(ex)
        sys.exit(ABNORMAL_EXIT_CODE)


@click.group()
def cli():
    pass


@cli.command()
@finetune_options()
def finetune(*args, **kwargs):
    """
    command line finetune entrance
    """
    record_operation_and_service_info_log('Start finetune.')
    finetune_task = FinetuneTask(*args, **kwargs)
    return finetune_task.start()


@cli.command()
@infer_options()
def infer(*args, **kwargs):
    """
    command line infer entrance
    """
    record_operation_and_service_info_log('Start infer.')
    infer_task = EvaluateInferTask(task_type=INFER_TASK_NAME, *args, **kwargs)
    return infer_task.start()


@cli.command()
@evaluate_options()
def evaluate(*args, **kwargs):
    """
    command line evaluate entrance
    """
    record_operation_and_service_info_log('Start evaluate.')
    evaluate_task = EvaluateInferTask(task_type=EVALUATE_TASK_NAME, *args, **kwargs)
    return evaluate_task.start()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved."""

import os
import subprocess
from mindpet.utils.io_utils import read_yaml_file
from mindpet.utils.constants import MODEL_CONFIG_ROOT_KEYS
from mindpet.utils.exceptions import LinkPathError, TaskError, ReadYamlFileError, UnexpectedError, CreateProcessError, \
    MonitorProcessRspError
from mindpet.log.log import record_operation_and_service_error_log, record_operation_and_service_info_log, \
    operation_logger_without_std
from mindpet.utils.task_utils import create_output_path_subdir_with_uuid, model_config_params_check_item, \
    extend_model_config_freeze_command, monitor_process_rsp_code, model_config_keys_check_item, path_in_out_loop_check


class FinetuneTask:
    """FinetuneTask class"""
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.command = []
        self.command_params = []

        try:
            self._process_param_and_command()
        except KeyboardInterrupt as ex:
            record_operation_and_service_error_log(
                'Finetune task is terminated by current user, task has stopped and exited.')
            raise ex
        except Exception as ex:
            record_operation_and_service_error_log('Finetune failed.')
            raise ex

    # pylint: disable=R1732
    def start(self):
        """
        启动命令
        :return: 微调执行状态
        """
        record_operation_and_service_info_log('Finetune task is running.')

        try:
            process = subprocess.Popen(self.command, env=os.environ, shell=False,
                                       stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        except KeyboardInterrupt as ex:
            record_operation_and_service_error_log('Finetune task is terminated by current user, '
                                                   'task has stopped and exited.')
            raise ex
        except Exception as ex:
            if ex is None or not str(ex):
                raise CreateProcessError('Exception occurred when creating finetune task process, '
                                         'no error message available.') from ex
            raise CreateProcessError(f'Exception occurred when creating finetune task process, '
                                     f'error message: {str(ex)}.') from ex

        try:
            rsp_code = monitor_process_rsp_code(task_name='finetune', task_process=process,
                                                timeout=self.kwargs.get('timeout'))
        except KeyboardInterrupt as ex:
            record_operation_and_service_error_log('Finetune task is terminated by current user, '
                                                   'task has stopped and exited.')
            raise ex
        except Exception as ex:
            if ex is None or not str(ex):
                raise MonitorProcessRspError('Exception occurred when monitoring finetune task process, '
                                             'no error message available.') from ex
            raise MonitorProcessRspError('Exception occurred when monitoring finetune task process, '
                                         'error message: {str(ex)}.') from ex

        if rsp_code != 0:
            operation_logger_without_std.error('Finetune failed.')
            raise TaskError('Finetune failed.')

        record_operation_and_service_info_log('Finetune successfully.')
        return True

    def _process_param_and_command(self):
        """
        调优任务启动前的准备工作
        """
        # 命令行参数准备
        self._prepare_command_params()

        # 命令准备
        self._prepare_command()

    def _prepare_command_params(self):
        """
        准备命令行参数
        """
        # 输入、输出回环路径判断
        path_in_out_loop_check(self.kwargs)

        self._wrap_optional_params()
        self._wrap_model_config_params()
        self._wrap_required_params()

    def _wrap_optional_params(self):
        """
        组装选填参数
        """
        pretrained_model_path = self.kwargs.get('pretrained_model_path')
        if pretrained_model_path is not None:
            self.command_params.extend(['--pretrained_model_path', str(pretrained_model_path)])

    def _wrap_required_params(self):
        """
        组装必填参数
        """
        output_path = create_output_path_subdir_with_uuid(self.kwargs.get('output_path'))

        self.command_params.extend(['--output_path', output_path])
        self.command_params.extend(['--data_path', self.kwargs.get('data_path')])

    def _wrap_model_config_params(self):
        """
        针对model_config配置文件内容，组装相关参数
        """
        model_config_option = self.kwargs.get('model_config_path')

        if model_config_option is not None:
            try:
                content = read_yaml_file(model_config_option)
            except (FileNotFoundError, LinkPathError) as ex:
                raise ReadYamlFileError(
                    f'Error occurred when getting finetune params info from param [model_config_path], '
                    f'error message: {str(ex)}') from ex
            except Exception as ex:
                raise UnexpectedError(
                    f'Unexpected error occurred when getting finetune params info from param [model_config_path], '
                    f'error message: {str(ex)}') from ex

            model_config_keys_check_item(content, MODEL_CONFIG_ROOT_KEYS)
            model_config_params_check_item(task_object=self, content=content)
            extend_model_config_freeze_command(content=content, task_object=self,
                                               model_config_path=self.kwargs.get('model_config_path'))

    def _prepare_command(self):
        """
        组装命令
        """
        self.command.append('python')
        self.command.append(self.kwargs.get('boot_file_path'))
        self.command.extend(self.command_params)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""Version APIs"""
def is_version_le(current_version, base_version):
    """
    description: Check whether the current version is lower than or equal to the base version.
    args:
        current_version: string, like "1.10.1"
        base_version: string, like "2.0.0rc1"
    return:
        Bool: current_version <= base_version
    example:
        for current_version: 1.10.1, base_version: 2.0.0rc1, it return True.
    """
    version_split_char = '.'
    if version_split_char not in base_version or version_split_char not in current_version:
        raise ValueError("The version string will contain the `.`."
                         "For example, current_version 1.8.1, base_version: 2.0.0.")
    for x, y in zip(current_version.split(version_split_char), base_version.split(version_split_char)):
        if not x.isdigit() or not y.isdigit():
            continue
        if int(x) != int(y):
            return int(x) <= int(y)
    return True


def is_version_ge(current_version, base_version):
    """
    description: Check whether the current version is higher than or equal to the base version.
    args:
        current_version: string, like "1.10.1"
        base_version: string, like "2.0.0rc1"
    return:
        Bool: current_version >= base_version
    example:
        for current_version: 1.10.1, base_version: 2.0.0rc1, it return False.
    """
    version_split_char = '.'
    if version_split_char not in base_version or version_split_char not in current_version:
        raise ValueError("The version string will contain the `.`."
                         "For example, current_version 1.8.1， base_version: 2.0.0.")
    for x, y in zip(current_version.split(version_split_char), base_version.split(version_split_char)):
        if not x.isdigit() or not y.isdigit():
            continue
        if int(x) != int(y):
            return int(x) >= int(y)
    return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""Constants for delta"""
from mindspore import dtype as mstype

# 算法精度与类型控制
VALID_TENSOR_DATATYPE = [mstype.float16, mstype.float32, mstype.bfloat16]



# Copyright 2023 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
p-tuning-v2
Reference: https://arxiv.org/pdf/2110.07602.pdf
"""

import mindspore as ms
import numpy as np
from mindspore import dtype as mstype
from mindspore import nn
from mindspore.common.initializer import initializer
from mindspore.ops import operations as P

try:
    from mindspore._checkparam import Rel, Validator

    INC_LEFT = Rel.INC_LEFT
except ImportError:
    import mindspore._checkparam as Validator

    INC_LEFT = Validator.INC_LEFT

from mindpet.utils.version_control import get_dropout


class PrefixEmbedding(nn.Cell):
    """
    The embedding with parallel_config.
    """

    def __init__(self, vocab_size, embedding_size, data_parallel=1, model_parallel=1, vocab_emb_dp=True,
                 param_init='normal'):
        """
        vocab_size: vocab size
        embedding_size: embedding size
        data_parallel: data parallel config
        model_parallel: data parallel config
        vocab_emb_dp: embedding dp config
        param_init: parameter init method
        """
        super().__init__()

        self.vocab_size = vocab_size
        self.embedding_size = embedding_size
        self.embedding_table = ms.Parameter(initializer(param_init, [self.vocab_size, self.embedding_size]),
                                            name='embedding_table', parallel_optimizer=False)
        if vocab_emb_dp:
            self.gather = P.Gather().shard(((1, 1), (data_parallel, 1)))
        else:
            if self.vocab_size % model_parallel != 0:
                raise ValueError(f"The vocab size of the embedding {self.vocab_size} must be a "
                                 f"multiple of model_parallel {model_parallel}.")
            self.gather = P.Gather().shard(((model_parallel, 1), (data_parallel, 1)))

    def construct(self, input_ids):
        """
        embedding inputs
        """
        output = self.gather(self.embedding_table, input_ids, 0)
        return output


class PrefixEncoder(nn.Cell):
    """
    The cell to encode the prefix
    Input : batch_size
    Output shape: layers * (2, bs, num_heads, pre_len, kv_channels)
    """

    def __init__(
            self,
            pre_seq_len,
            num_layers,
            num_heads,
            kv_channels,
            prefix_projection,
            projection_dim,
            dropout_prob,
            parallel_config=None,
            out_perm=(2, 0, 3, 1, 4),
    ):
        """
        Args:
            pre_seq_len: prefix的序列长度
            num_layers: 原模型transformer层数
            num_heads: 原模型transformer多头注意力头数
            kv_channels: 原模型transformer kv维度
            prefix_projection 是否使用MLP表征
            projection_dim: MLP维度
            dropout_prob: 丢弃率
            parallel_config: 并行参数
            out_perm: 输出的维度
        """
        super().__init__()
        self.pre_seq_len = Validator.check_positive_int(pre_seq_len, "pre_seq_len")
        self.num_layers = Validator.check_positive_int(num_layers, "num_layers")
        self.num_heads = Validator.check_positive_int(num_heads, "num_heads")
        self.kv_channels = Validator.check_positive_int(kv_channels, "kv_channels")

        dropout_prob = Validator.check_float_range(dropout_prob, 0.0, 1.0, INC_LEFT)
        self.dropout = get_dropout(dropout_prob)

        self.prefix_projection = prefix_projection

        self.mindpet_delta_ptuning2_prefix = ms.Parameter(
            np.arange(self.pre_seq_len), requires_grad=False
        )

        out_embed_dim = self.num_layers * self.kv_channels * self.num_heads * 2
        if parallel_config:
            data_parallel = Validator.check_positive_int(parallel_config.data_parallel, "data_parallel")
            model_parallel = Validator.check_positive_int(parallel_config.model_parallel, "model_parallel")
            vocab_emb_dp = parallel_config.vocab_emb_dp
        else:
            data_parallel = 1
            model_parallel = 1
            vocab_emb_dp = True

        self.mindpet_delta_ptuning2_embedding = PrefixEmbedding(vocab_size=self.pre_seq_len,
                                                                embedding_size=out_embed_dim,
                                                                data_parallel=data_parallel,
                                                                model_parallel=model_parallel,
                                                                vocab_emb_dp=vocab_emb_dp)

        if self.prefix_projection:
            self.projection_dim = Validator.check_positive_int(
                projection_dim, "projection_dim"
            )
            # two-layer MLP to encode the prefix
            self.mindpet_delta_ptuning2_dense_in = nn.Dense(out_embed_dim, self.projection_dim)
            self.mindpet_delta_ptuning2_tanh = nn.Tanh()
            self.mindpet_delta_ptuning2_dense_out = nn.Dense(self.projection_dim, out_embed_dim)
            self.mindpet_delta_ptuning2_trans = nn.SequentialCell(
                self.mindpet_delta_ptuning2_dense_in,
                self.mindpet_delta_ptuning2_tanh,
                self.mindpet_delta_ptuning2_dense_out,
            )

        self.out_perm = out_perm
        self.expand_dims = P.ExpandDims().shard(((1,),))
        self.tile = P.Tile().shard(((1, 1),))
        self.cast = P.Cast()
        self.reshape = P.Reshape()
        self.transpose = P.Transpose()
        self.spilt_layers = P.Split(axis=0, output_num=self.num_layers)

    def construct(self, batch_size, dtype=mstype.half):
        """
        new prefix
        """
        prefix_tokens = self.expand_dims(self.mindpet_delta_ptuning2_prefix, 0)
        prefix_tokens = self.tile(prefix_tokens, (batch_size, 1))

        # (bs, pre_len) -> (bs, pre_len, 2 * layers * num_heads * kv_channels)
        past_key_values = self.mindpet_delta_ptuning2_embedding(prefix_tokens)

        if self.prefix_projection:
            past_key_values = self.mindpet_delta_ptuning2_trans(past_key_values)

        past_key_values = self.cast(past_key_values, dtype)

        # (bs, pre_len, 2 * layers * num_heads * kv_channels) -> (bs, pre_len, 2 * layers, num_heads, kv_channels)
        past_key_values = self.reshape(past_key_values, (batch_size,
                                                         self.pre_seq_len,
                                                         self.num_layers * 2,
                                                         self.num_heads,
                                                         self.kv_channels,))

        past_key_values = self.dropout(past_key_values)

        # (bs, pre_len, 2 * layers, num_heads, kv_channels) -> (2 * layers, bs, num_heads, pre_len, kv_channels)
        past_key_values = self.transpose(past_key_values, self.out_perm)

        # (2 * layers, bs, num_heads, pre_len, kv_channels) -> , layers * (2, bs, num_heads, pre_len, kv_channels)
        past_key_values = self.spilt_layers(past_key_values)

        return past_key_values

    def shard(self, data_parallel, model_parallel):
        """
        set shard strategy
        """

        if self.prefix_projection:
            # (bs, pre_len, embedding_dim)
            self.mindpet_delta_ptuning2_dense_in.matmul.shard(((data_parallel, 1), (model_parallel, 1)))
            self.mindpet_delta_ptuning2_dense_in.bias_add.shard(((data_parallel, 1), (1,)))
            self.mindpet_delta_ptuning2_tanh.tanh.shard(((data_parallel, 1, 1),))
            self.mindpet_delta_ptuning2_dense_out.matmul.shard(((data_parallel, 1), (model_parallel, 1)))
            self.mindpet_delta_ptuning2_dense_out.bias_add.shard(((data_parallel, 1), (1,)))

        # (bs, pre_len, 2 * layers * num_heads * kv_channels)
        self.cast.shard(((data_parallel, 1, 1),))  # (dp, 1, 1)

        # (bs, pre_len, 2 * layers, num_heads, kv_channels)
        self.dropout.dropout.shard(((data_parallel, 1, 1, 1, 1),))  # (dp, 1, 1, 1, 1)
        self.transpose.shard(((data_parallel, 1, 1, 1, 1),))  # (dp, 1, 1, 1, 1)

        # (2 * layers, bs, num_heads, pre_len, kv_channels)
        self.spilt_layers.shard(((1, data_parallel, 1, 1, 1),))  # (1, dp, 1, 1, 1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""path check param"""

from mindpet.utils.constants import SPACE_CHARACTER
from mindpet.security.param_check.option_check_utils import PathLengthCheckParam, PathContentCheckParam
from mindpet.utils.constants import DEFAULT_PATH_LEN_MIN_LIMIT, DEFAULT_PATH_LEN_MAX_LIMIT, DEFAULT_FILE_LEN_MIN_LIMIT, \
    DEFAULT_FILE_LEN_MAX_LIMIT

default_folder_length_check_param = PathLengthCheckParam(path_min_limit=DEFAULT_PATH_LEN_MIN_LIMIT,
                                                         path_max_limit=DEFAULT_PATH_LEN_MAX_LIMIT,
                                                         file_min_limit=None,
                                                         file_max_limit=None)
default_file_length_check_param = PathLengthCheckParam(path_min_limit=DEFAULT_PATH_LEN_MIN_LIMIT,
                                                       path_max_limit=DEFAULT_PATH_LEN_MAX_LIMIT,
                                                       file_min_limit=DEFAULT_FILE_LEN_MIN_LIMIT,
                                                       file_max_limit=DEFAULT_FILE_LEN_MAX_LIMIT)
default_folder_content_check_param = PathContentCheckParam(base_whitelist_mode='ALL',
                                                           extra_whitelist=['/', '-', '_', SPACE_CHARACTER])
default_file_content_check_param = PathContentCheckParam(base_whitelist_mode='ALL',
                                                         extra_whitelist=['.', '/', '-', '_', SPACE_CHARACTER])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""Set up"""
from distutils.core import setup
from setuptools import find_packages

with open('README.md', encoding='utf-8') as f:
    readme = f.read()

cmd_class = {}


def get_version():
    version = '1.0.3'
    return version


def do_setup(packages_data):
    setup(
        name='mindpet',

        version=get_version(),

        description='Parameter-Efficient Tuning',

        keywords='Parameter-Efficient Tuning',

        # 详细的程序描述
        long_description=readme,
        long_description_content_type='text/markdown',

        # 编包依赖
        setup_requires=[
            'python_version>=3.7', 'setuptools>=61.2.0', 'pyyaml>=6.0', 'wheel'
        ],

        # 项目依赖的Python库
        install_requires=[
            'click', 'pyyaml'
        ],

        # 需要打包的目录列表
        packages=find_packages(
            # 不需要打包的目录列表
            exclude=[
                'test', 'test.*'
            ]
        ),

        package_data=packages_data,
        entry_points={
            'console_scripts': [
                'mindpet = mindpet.mindpet_main:cli_wrapper'
            ],
        },
        cmdclass=cmd_class
    )


if __name__ == '__main__':
    package_data = {
        'mindpet': [
            '*.py',
            '*/*.py',
            '*/*/*.py',
            '*/*/*/*.py',
            '*/*/*/*/*.py'
        ]
    }
    do_setup(package_data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Huawei Technologies Co., Ltd. 2022-2023. All rights reserved.
"""boot file path option"""
import os
import click
from click.exceptions import MissingParameter, BadParameter
from mindpet.utils.constants import PATH_MODE_LIMIT
from mindpet.task.finetune.finetune_options_check import FinetuneOptionsCheckParam, FinetuneOptionsCheck
from mindpet.task.options.path_check_param import default_file_length_check_param, default_file_content_check_param


class BootFilePathOption(click.core.Option):
    """BootFilePathOption class"""
    def __init__(self):
        super().__init__(
            param_decls=('-bp', '--boot_file_path'),
            type=str,
            help='boot file local path',
            callback=self.boot_file_path_callback
        )

    @staticmethod
    def _is_py_file(path):
        """
        路径是否是python文件

        :param path: 指定路径
        :return: 路径是否是python文件
        """
        return os.path.isfile(path) and path.endswith('.py')

    def boot_file_path_callback(self, ctx, param, value):
        """
        boot_file_path参数click回调方法

        :param ctx: 上下文信息
        :param param: 参数属性
        :param value: 输入值
        :return: 回调处理后的参数值
        """
        if value is None:
            raise MissingParameter('Param [boot_file_path] is required.')

        check_param = FinetuneOptionsCheckParam(path_length_check_param=default_file_length_check_param,
                                                path_content_check_param=default_file_content_check_param,
                                                mode=PATH_MODE_LIMIT,
                                                path_including_file=True,
                                                force_quit=True,
                                                quiet=ctx.params.get('quiet', False))
        boot_file_path_option = FinetuneOptionsCheck(option_name=param.name, option_value=value)
        boot_file_path_option.check(check_param)

        if not self._is_py_file(boot_file_path_option.option_value):
            raise BadParameter('Param [boot_file_path] only support .py file path.')

        return boot_file_path_option.option_value
